
/*################################################################################################
 *##################                                                            ##################
 *##################                     INTERNAL TOOLS                         ##################
 *##################                                                            ##################
 *##################                                                            ##################
 *##################                     Implementation                         ##################
 *##################                                                            ##################
 *################################################################################################
 */

/*
 * OKit:   Object Tree,  RPL Virtual Machine  and  compiler  shared library
 *
 * Copyright (C) 2000 Yann LANDRIN-SCHWEITZER
 *
 *
 * This library is free software;  you can redistribute it and/or modify it
 * under  the terms of the  GNU Lesser General Public License  as published
 * by the  Free Software Foundation;  either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This  library  is  distributed  in  the hope that it will be useful, but
 * WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty  of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library;  if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * Author:   Yann LANDRIN-SCHWEITZER
 * Contact:  landrin@multimania.com
 * Homepage: http://landrin.multimania.com/
 */


#ifndef _OKIT_INTERNAL_
#define _OKIT_INTERNAL_


//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Printing and scanning
/////


// Function _printinteger
// ======================
// 
// Prints an unsigned long integer into a string.
// 
//   Args:
//         _u             : integer to print
//         _s             : string to print into
//         _m             : number of characters to print
//         _b             : base
// 
//   Returns:
//         _0             : if something went wrong
// 	otherwise the number of characters printed

unsigned long _printinteger(unsigned long u, char * s, unsigned long m, unsigned long b) ;


// Function _scaninteger
// =====================
// 
// Scans a string for an unsigned long integer.
// 
//   Args:
//         _u             : pointer to an integer that will receive the scaned value
//         _s             : string to scan from
//         _m             : number of characters to scan
//         _b             : base
// 
//   Returns:
//         _0             : if something went wrong
// 	otherwise the number of characters scanned

unsigned long _scaninteger(unsigned long * u, char * s, unsigned long m, unsigned long b) ;


// Function _printfloat
// ====================
// 
// Prints a long double into a string.
// 
//   Args:
//         _u             : double to print
//         _s             : string to print into
//         _m             : number of characters to print
//         _e             : precision
// 
//   Returns:
//         _0             : if something went wrong
// 	otherwise the number of characters printed

unsigned long _printfloat(long double u, char * s, unsigned long m, unsigned long e) ;


// Function _scanfloat
// ===================
// 
// Scans a string for a long double.
// 
//   Args:
//         _u             : pointer to a double will receive the scaned value
//         _s             : string to scan from
//         _m             : number of characters to scan
// 
//   Returns:
//         _0             : if something went wrong
// 	otherwise the number of characters scanned

unsigned long _scanfloat(long double * u, char * s, unsigned long m) ;


//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Search and key lookup
/////


// Function _keyncmp
// =================
// 
// Compares lexically two keys on a given number of characters.
// 
//   Args:
//         _ky1           : first key
//         _ky2           : second key
//         _len           : maximum number of characters to take into account
// 
//   Returns:
//         _ 0            : if the two keys where the same up to 'len' chars
//         _-1            : if the first key was found smaller than the second
//         _+1            : if the first key was found greater than the second

signed char _keyncmp(char * ky1, char * ky2, unsigned long len) ;


// Function _keysearch
// ===================
// 
// Finds the position of a key in an ordered array of keys.
// 
//   Args:
//         _key           : the key to look up
//         _keys          : the arrays of keys
//         _keynum        : the size of the 'keys' table
//         _keylen        : the maximum length of keys
//         _pos           : pointer towards an integer to fill with the position
//                            of the first greater or equal key
// 
//   Returns:
//         _-1            : if the key was not found
//         otherwise the index of the given key in the table.
//
//   Note:
//         _ The key array must be ordered in growing lexical order for this function to work.
//         _ When -1 is returned, *pos contains the index of the first greater key found,
//               that is: 0 if the key was smaller than the first in the array, keynum if
//               it was greater that the last in the array, otherwise the index of a real key.
//         _ When the key was found, *pos also holds the returned index value.

unsigned long _keysearch(char * key, char * * keys, unsigned long keynum, unsigned long keylen, unsigned long * pos) ;




#endif //_OKIT_INTERNAL_
