
/*################################################################################################
 *##################                                                            ##################
 *##################                    Execution module                        ##################
 *##################                                                            ##################
 *##################                      Header  file                          ##################
 *##################                                                            ##################
 *################################################################################################
 */

/*
 * OKit:   Object Tree,  RPL Virtual Machine  and  compiler  shared library
 *
 * Copyright (C) 2000 Yann LANDRIN-SCHWEITZER
 *
 *
 * This library is free software;  you can redistribute it and/or modify it
 * under  the terms of the  GNU Lesser General Public License  as published
 * by the  Free Software Foundation;  either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This  library  is  distributed  in  the hope that it will be useful, but
 * WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty  of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library;  if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * Author:   Yann LANDRIN-SCHWEITZER
 * Contact:  landrin@multimania.com
 * Homepage: http://landrin.multimania.com/
 */



#ifndef _OKIT_EXECMOD_

#define _OKIT_EXECMOD_

#include <OKit/ObjTypes.h>
#include <OKit/SysTypes.h>



//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Standard error codes and messages
/////


#define STAT_NAME_LEN	64

struct S_Status {
	unsigned long num ;
	char * msg ;
	char com[STAT_NAME_LEN+1] ;
} ;

typedef struct S_Status Status ;

typedef Status (* CodeOperator) (Stack stk, NameSpace dom, Object data) ;

typedef Status (* BltnOperator) (Stack stk, NameSpace dom) ;


#define OK		 0
#define UNHDLD		 1
#define RNTERR		 2
#define DTINCR		 3
#define SYSERR		 4
#define IOSERR		 5
#define EXCERR		 6
#define NOBLTN		 7
#define UNIMPL		 8
#define NOUSER		 9
#define TOOFEWARG 	10
#define INVARGTYP 	11
#define INVARGVAL	12
#define M13		13
#define M14		14
#define M15		15
#define M16		16
#define M17		17
#define M18		18
#define M19		19
#define MAXSYSERR	20
#define INTERNERR	~0

extern char * ExecError[] ;


//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Object evaluation
/////


Status ObjectEval(Stack stk, NameSpace dom, Object obj) ;

Status ListEval(Stack stk, NameSpace dom, List lst) ;

Status CodeEval(Stack stk, NameSpace dom, Code cod) ;

Status BltnEval(Stack stk, NameSpace dom, Bltn blt) ;


#endif // _OKIT_EXECMOD_
