
/*################################################################################################
 *##################                                                            ##################
 *##################                      Object Types                          ##################
 *##################                                                            ##################
 *##################                      Header  file                          ##################
 *##################                                                            ##################
 *################################################################################################
 */

/*
 * OKit:   Object Tree,  RPL Virtual Machine  and  compiler  shared library
 *
 * Copyright (C) 2000 Yann LANDRIN-SCHWEITZER
 *
 *
 * This library is free software;  you can redistribute it and/or modify it
 * under  the terms of the  GNU Lesser General Public License  as published
 * by the  Free Software Foundation;  either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This  library  is  distributed  in  the hope that it will be useful, but
 * WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty  of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library;  if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * Author:   Yann LANDRIN-SCHWEITZER
 * Contact:  landrin@multimania.com
 * Homepage: http://landrin.multimania.com/
 */



#ifndef _OKIT_OBJTYPES_

#define _OKIT_OBJTYPES_



//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Object types specifications
/////

typedef unsigned short Type ;


struct S_Object {
	Type type ;	// Object type
	unsigned long refc ;	// Reference count (for garbage collection)
	unsigned long flag ;	// Internal flags
	unsigned long size ;	// Object size
	void * obj ;		// Object pointer
} ;

typedef struct S_Object * Object ;

#define OT_VOID		0x00
#define OT_INTG		0x01
#define OT_FLTP		0x02
#define OT_STRG		0x04
#define OT_LIST		0x08
#define OT_CODE		0x10
#define OT_BLTN		0x20

// VOID Object Type
typedef void _Void ;
typedef void * Void ;
#define ST_VOID		0

// INTG Object Type
typedef long _Intg ;
typedef long * Intg ;
#define ST_INTG		sizeof(long)

// FLTP Object Type
typedef double _Fltp ;
typedef double * Fltp ;
#define ST_FLTP		sizeof(double)

// STRG Object Type
struct S_Strg {
	unsigned long len ;
	char * chr ;
} ;
typedef struct S_Strg _Strg ;
typedef struct S_Strg * Strg ;
#define ST_STRG		sizeof(struct S_Strg)

// LIST Object Type
struct S_Node {
	struct S_Node * last ;
	struct S_Node * next ;
	struct S_Object * obj ;
} ;
struct S_List {
	struct S_Node * frst ;
	struct S_Node * last ;
} ;
typedef struct S_List _List ;
typedef struct S_List * List ;
#define ST_LIST		sizeof(struct S_List)

// CODE Object Type
#define CODE_NAME_LEN	64
#define CODE_ARGT_LEN	 8
struct S_Code {
	unsigned char nmcl ;
	char name[CODE_NAME_LEN+1] ;
	unsigned char argc ;
	Type argt[CODE_ARGT_LEN] ;
	unsigned char retc ;
	void * func ;
	struct S_Object * data ;
} ;
typedef struct S_Code _Code ;
typedef struct S_Code * Code ;
#define ST_CODE		sizeof(struct S_Code)

// BLTN Object Type
#define BLTN_NAME_LEN	64
#define BLTN_ARGT_LEN	 8
struct S_Bltn {
	char name[BLTN_NAME_LEN+1] ;
	unsigned char argc ;
	Type argt[BLTN_ARGT_LEN] ;
	unsigned char retc ;
	void * func ;
	char * desc ;
} ;
typedef struct S_Bltn _Bltn ;
typedef struct S_Bltn * Bltn ;
#define ST_BLTN		sizeof(struct S_Bltn)



//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Internal object types manipulation
/////


// Generic Object

Object ObjectAlloc() ;
Object ObjectCopy(Object obj) ;
Object ObjectClone(Object obj) ;
Object ObjectDup(Object obj) ;
Object ObjectDrop(Object obj) ;
void ObjectFree(Object obj) ;
Type ObjectType(Object obj) ;
char * ObjectTypeName(Type typ) ;
void * ObjectContent(Object obj) ;
unsigned long ObjectStat(Object obj) ;
unsigned long ObjectOut(Object obj, char * out, unsigned long max) ;
unsigned char ObjectEqual(Object obj, Object objw) ;
Strg ObjectStrg(Object obj) ;


// Void

Void VoidNew() ;
Object VoidWrap(Void vd) ;
unsigned long VoidOut(Void vd, char * out, unsigned long max) ;


// Intg

Intg IntgAlloc() ;
Intg IntgCopy(Intg in) ;
Intg IntgClone(Intg in) ;
Intg IntgNew(long nb) ;
Object IntgWrap(Intg in) ;
unsigned long IntgOut(Intg in, char * out, unsigned long max) ;
unsigned char IntgEqual(Intg in, Intg inw) ;


// Fltp

Fltp FltpAlloc() ;
Fltp FltpCopy(Fltp fp) ;
Fltp FltpClone(Fltp fp) ;
Fltp FltpNew(double nb) ;
Object FltpWrap(Fltp fp) ;
unsigned long FltpOut(Fltp fp, char * out, unsigned long max) ;
unsigned char FltpEqual(Fltp fp, Fltp fpw) ;


// Strg

Strg StrgAlloc(unsigned long size) ;
Strg StrgCopy(Strg str) ;
Strg StrgClone(Strg str) ;
void StrgFree(Strg str) ;
Strg StrgNew(char * chs) ;
Object StrgWrap(Strg str) ;
unsigned long StrgLen(Strg str) ;
Strg StrgCat(Strg str1, Strg str2) ;
Strg StrgCut(Strg str, unsigned long beg, unsigned long end) ;
unsigned long StrgOut(Strg str, char * out, unsigned long max) ;
unsigned char StrgEqual(Strg str, Strg strw) ;
unsigned long StrgCharString(Strg str, char * out, unsigned long max) ;

// List

List ListAlloc() ;
List ListAllocBlock(unsigned long numb) ;
List ListCopy(List lst) ;
List ListClone(List lst) ;
List ListDup(List lst) ;
List ListDrop(List lst) ;
void ListFree(List lst) ;
List ListNew() ;
Object ListWrap(List lst) ;
unsigned long ListLen(List lst) ;
Object ListGet(List lst, unsigned long pos) ;
Object ListPut(List lst, unsigned long pos, Object obj) ;
List ListDel(List lst, unsigned long beg, unsigned long end) ;
List ListIns(List lst, unsigned long pos, List lsi) ;
List ListPush(List lst, Object obj) ;
Object ListPop(List lst) ;
List ListUnshift(List lst, Object obj) ;
Object ListShift(List lst) ;
unsigned long ListOut(List lst, char * out, unsigned long max) ;
unsigned char ListEqual(List lst, List lstw) ;


// Code
Code CodeAlloc() ;
Code CodeCopy(Code cod) ;
Code CodeClone(Code cod) ;
Code CodeDup(Code cod) ;
Code CodeDrop(Code cod) ;
void CodeFree(Code cod) ;
Code CodeNew() ;
Object CodeWrap(Code cod) ;
unsigned char CodeArgC(Code cod) ;
Type CodeArgT(Code cod, unsigned char argn) ;
unsigned char CodeRetC(Code cod) ;
unsigned char CodeSetArgC(Code cod, unsigned char argc) ;
Type CodeSetArgT(Code cod, unsigned char argn, Type argt) ;
unsigned char CodeSetRetC(Code cod, unsigned char retc) ;
char * CodeName(Code cod) ;
Strg CodeStrgName(Code cod) ;
void * CodeGetFunc(Code cod) ;
Object CodeGetData(Code cod) ;
void * CodeSetFunc(Code cod, void * func) ;
Object CodeSetData(Code cod, Object data) ;
unsigned long CodeOut(Code cod, char * out, unsigned long max) ;
unsigned char CodeEqual(Code cod, Code codw) ;


// Bltn
Bltn BltnAlloc() ;
Bltn BltnCopy(Bltn blt) ;
Bltn BltnClone(Bltn blt) ;
void BltnFree(Bltn blt) ;
Bltn BltnNew() ;
Object BltnWrap(Bltn blt) ;
unsigned char BltnArgC(Bltn blt) ;
Type BltnArgT(Bltn blt, unsigned char argn) ;
unsigned char BltnRetC(Bltn blt) ;
char * BltnName(Bltn blt) ;
Strg BltnStrgName(Bltn blt) ;
void * BltnFunc(Bltn blt) ;
unsigned long BltnOut(Bltn blt, char * out, unsigned long max) ;
unsigned char BltnEqual(Bltn blt, Bltn bltw) ;






#endif // _OKIT_OBJTYPES_

