
/*################################################################################################
 *##################                                                            ##################
 *##################                      System Types                          ##################
 *##################                                                            ##################
 *##################                      Header  file                          ##################
 *##################                                                            ##################
 *################################################################################################
 */

/*
 * OKit:   Object Tree,  RPL Virtual Machine  and  compiler  shared library
 *
 * Copyright (C) 2000 Yann LANDRIN-SCHWEITZER
 *
 *
 * This library is free software;  you can redistribute it and/or modify it
 * under  the terms of the  GNU Lesser General Public License  as published
 * by the  Free Software Foundation;  either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This  library  is  distributed  in  the hope that it will be useful, but
 * WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty  of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library;  if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * Author:   Yann LANDRIN-SCHWEITZER
 * Contact:  landrin@multimania.com
 * Homepage: http://landrin.multimania.com/
 */



#ifndef _OKIT_SYSTYPES_

#define _OKIT_SYSTYPES_

#include <OKit/ObjTypes.h>



//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Stack definition
/////

#define DEFCHNKSIZE 1 ;

struct S_Stack {
	unsigned long size ;
	unsigned long chnk ;
	unsigned long dpth ;
	Object * data ;
} ;

typedef struct S_Stack * Stack ;

Stack StackAlloc(unsigned long chksz) ;
void StackFree(Stack stk) ;
unsigned long StackDepth(Stack stk) ;
unsigned long StackPush(Stack stk, Object obj) ;
Object StackPop(Stack stk) ;
Object StackGet(Stack stk, unsigned long num) ;
Object StackPut(Stack stk, Object obj, unsigned long num) ;
unsigned long StackChkType(Stack stk, unsigned long argc, Type * argt) ;
unsigned long StackOut(Stack stk, char * out, unsigned long max, char * head, char * rowfmt, char * tail, unsigned long rows) ;
unsigned long StackStat(Stack stk) ;



//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Variables domain definition
/////


#define NAMELEN 128

struct S_NameSpace {
	unsigned long size ;
	unsigned long chnk ;
	unsigned long dpth ;
	char * * names ;
	Object * conts ;
} ;

typedef struct S_NameSpace * NameSpace ;


NameSpace NameSpaceAlloc(unsigned long chksz) ;
void NameSpaceFree(NameSpace nms) ;
Object NameSpaceGet(NameSpace nms, char * name) ;
Object NameSpaceSet(NameSpace nms, char * name, Object obj) ;
Object NameSpaceDel(NameSpace nms, char * name) ;
List NameSpaceList(NameSpace nms) ;
unsigned long NameSpaceOut(NameSpace nms, char * out, unsigned long max, char * head, char * rowfmt, char * tail) ;
unsigned long NameSpaceStat(NameSpace nms) ;



#endif //_OKIT_SYSTYPES_

