
/*################################################################################################
 *##################                                                            ##################
 *##################                      ERROR TRACER                          ##################
 *##################                                                            ##################
 *##################                                                            ##################
 *##################                     Implementation                         ##################
 *##################                                                            ##################
 *################################################################################################
 */

/*
 * OKit:   Object Tree,  RPL Virtual Machine  and  compiler  shared library
 *
 * Copyright (C) 2000 Yann LANDRIN-SCHWEITZER
 *
 *
 * This library is free software;  you can redistribute it and/or modify it
 * under  the terms of the  GNU Lesser General Public License  as published
 * by the  Free Software Foundation;  either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This  library  is  distributed  in  the hope that it will be useful, but
 * WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty  of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library;  if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * Author:   Yann LANDRIN-SCHWEITZER
 * Contact:  landrin@multimania.com
 * Homepage: http://landrin.multimania.com/
 */



#include <stdlib.h>
#include <errno.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>

#include <Internal.h>
#include <OKit/ErrLog.h>

#define ERRLOG_DEPTH 1024
#define ERRLOG_NAME_LEN 32
#define ERRLOG_MESS_LEN 128
#define ERRFATAL_DEPTH 64





//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Error Log
/////

unsigned long ERRLOGdbgmsk = ~0 ;                   // Current debugging mask
unsigned long ERRLOGcount = 0 ;                     // Total error count
unsigned long ERRLOGmark = 0 ;                      // Mark to first error in block
unsigned long ERRLOGent[ERRLOG_DEPTH] ;             // Error entry
unsigned long ERRLOGnum[ERRLOG_DEPTH] ;             // Error numbers
unsigned long ERRLOGsec[ERRLOG_DEPTH] ;             // Seconds
unsigned long ERRLOGmsc[ERRLOG_DEPTH] ;             // Microseconds
unsigned char ERRLOGlev[ERRLOG_DEPTH] ;             // Debug level
char ERRLOGname[ERRLOG_DEPTH][ERRLOG_NAME_LEN+1] ;  // Function names
char ERRLOGmess[ERRLOG_DEPTH][ERRLOG_MESS_LEN+1] ;  // Messages
unsigned long ERRLOGwrap[ERRLOG_DEPTH] ;            // Error blocks
int ERRLOGout = 2 ;                                 // Out file

unsigned long ERRLOGfatals = 0 ;
unsigned long ERRLOGfatal[ERRFATAL_DEPTH] ;


void SETERRLOGFILE(int fd) {
	ERRLOGout = fd ;
}

void ERRTRACEON() {
	ERRLOGdbgmsk = ~0 ;
}

void ERRTRACEOFF() {
	ERRLOGdbgmsk = 0 ;
}

void ERRTRACEDBG(unsigned long dbgmsk) {
	ERRLOGdbgmsk = dbgmsk ;
}

void FATALERR(int err) {
	if((err!=0)&&(ERRLOGfatals<ERRFATAL_DEPTH)) {
		ERRLOGfatal[ERRLOGfatals] = err ;
		ERRLOGfatals ++ ;
	}
}

void ERRMARK() {
	ERRLOGmark = ERRLOGcount + 1 ;
}

void DEBUG(char * funcname, unsigned char dbglevel, char * message, ...) {
	va_list ap ;
	struct timeval tm ;
	register unsigned long n ;
	register unsigned long i ;
	register unsigned long j ;
	register unsigned long d ;
	unsigned long l ;
	char * s ;
	unsigned long u ;
	char c ;
	va_start(ap,message) ;
	gettimeofday(&tm,NULL) ;
	n = ERRLOGcount%ERRLOG_DEPTH ;
	ERRLOGent[n] = ERRLOGcount ;
	ERRLOGnum[n] = errno ;
	ERRLOGsec[n] = tm.tv_sec ;
	ERRLOGmsc[n] = tm.tv_usec ;
	ERRLOGlev[n] = (dbglevel>31)?31:dbglevel ;
	i = 0 ;
	while((i<ERRLOG_NAME_LEN)&&(funcname[i]!='\0')) {
		ERRLOGname[n][i] = funcname[i] ;
		i ++ ;
	}
	ERRLOGname[n][i] = '\0' ;
	if(message!=NULL) {
		i = 0 ;
		j = 0 ;
		while((i<ERRLOG_MESS_LEN)&&(message[j]!='\0')) {
			while((i<ERRLOG_MESS_LEN)&&(message[j]!='\0')&&(message[j]!='%')) {
				ERRLOGmess[n][i] = message[j] ;
				i ++ ;
				j ++ ;
			}
			if(message[j]=='%') {
				j ++ ;
				if(message[j]=='\0') break ;
				l = 0 ;
				if( (message[j]>='0')&&(message[j]<='9') ) {
					j += _scaninteger(&l,message+j,10,10) ;
					if(l>(ERRLOG_MESS_LEN-i)) l = ERRLOG_MESS_LEN-i ;
				}
				else l = ERRLOG_MESS_LEN-i ;
				switch(message[j]) {
					case '%':
						if(i<ERRLOG_MESS_LEN) {
							ERRLOGmess[n][i] = '%' ;
							i ++ ;
						}
						break ;
					case 'c':
						c = (char) va_arg(ap, int) ;
						if(i<ERRLOG_MESS_LEN) {
							ERRLOGmess[n][i] = c ;
							i ++ ;
						}
						break ;
					case 's':
						s = (char *) va_arg(ap, char *) ;
						if(s==NULL) break ;
						d = 0 ;
						while( (i<ERRLOG_MESS_LEN)&&(d<l)&&(s[d]!='\0') ) {
							ERRLOGmess[n][i] = s[d] ;
							i ++ ;
							d ++ ;
						}
						break ;
					case 'u':
						u = (unsigned long) va_arg(ap, int) ;
						i += _printinteger(u,ERRLOGmess[n]+i,l,10) ;
						break ;
					case 'x':
						u = (unsigned long) va_arg(ap, int) ;
						i += _printinteger(u,ERRLOGmess[n]+i,l,16) ;
						break ;
					default:
						;
				}
				j ++ ;
			}
		}
		ERRLOGmess[n][i] = '\0' ;
	}
	else {
		ERRLOGmess[n][0] = '\0' ;
	}
	ERRLOGwrap[n] = ERRLOGmark ;
	ERRLOGcount ++ ;

	i = 0 ;
	while(i<ERRLOGfatals) {
		if(ERRLOGfatal[i]==errno) break ;
		i ++ ;
	}

	if(i<ERRLOGfatals) {
		ERRLOGmark = ERRLOGcount ;
	}

	if(ERRLOGdbgmsk&(((unsigned long)0x1)<<ERRLOGlev[n])) ERRPRINT() ;

	if(i<ERRLOGfatals) {
		exit(errno) ;
	}

	va_end(ap) ;
}

void ERROR(char * funcname, char * message, ...) {
	va_list ap ;
	struct timeval tm ;
	register unsigned long n ;
	register unsigned long i ;
	register unsigned long j ;
	register unsigned long d ;
	unsigned long l ;
	char * s ;
	unsigned long u ;
	char c ;
	va_start(ap,message) ;
	gettimeofday(&tm,NULL) ;
	n = ERRLOGcount%ERRLOG_DEPTH ;
	ERRLOGent[n] = ERRLOGcount ;
	ERRLOGnum[n] = errno ;
	ERRLOGsec[n] = tm.tv_sec ;
	ERRLOGmsc[n] = tm.tv_usec ;
	ERRLOGlev[n] = 31 ;
	i = 0 ;
	while((i<ERRLOG_NAME_LEN)&&(funcname[i]!='\0')) {
		ERRLOGname[n][i] = funcname[i] ;
		i ++ ;
	}
	ERRLOGname[n][i] = '\0' ;
	if(message!=NULL) {
		i = 0 ;
		j = 0 ;
		while((i<ERRLOG_MESS_LEN)&&(message[j]!='\0')) {
			while((i<ERRLOG_MESS_LEN)&&(message[j]!='\0')&&(message[j]!='%')) {
				ERRLOGmess[n][i] = message[j] ;
				i ++ ;
				j ++ ;
			}
			if(message[j]=='%') {
				j ++ ;
				if(message[j]=='\0') break ;
				l = 0 ;
				if( (message[j]>='0')&&(message[j]<='9') ) {
					j += _scaninteger(&l,message+j,10,10) ;
					if(l>(ERRLOG_MESS_LEN-i)) l = ERRLOG_MESS_LEN-i ;
				}
				else l = ERRLOG_MESS_LEN-i ;
				switch(message[j]) {
					case '%':
						if(i<ERRLOG_MESS_LEN) {
							ERRLOGmess[n][i] = '%' ;
							i ++ ;
						}
						break ;
					case 'c':
						c = (char) va_arg(ap, int) ;
						if(i<ERRLOG_MESS_LEN) {
							ERRLOGmess[n][i] = c ;
							i ++ ;
						}
						break ;
					case 's':
						s = (char *) va_arg(ap, char *) ;
						if(s==NULL) break ;
						d = 0 ;
						while( (i<ERRLOG_MESS_LEN)&&(d<l)&&(s[d]!='\0') ) {
							ERRLOGmess[n][i] = s[d] ;
							i ++ ;
							d ++ ;
						}
						break ;
					case 'u':
						u = (unsigned long) va_arg(ap, int) ;
						i += _printinteger(u,ERRLOGmess[n]+i,l,10) ;
						break ;
					case 'x':
						u = (unsigned long) va_arg(ap, int) ;
						i += _printinteger(u,ERRLOGmess[n]+i,l,16) ;
						break ;
					default:
						;
				}
				j ++ ;
			}
		}
		ERRLOGmess[n][i] = '\0' ;
	}
	else {
		ERRLOGmess[n][0] = '\0' ;
	}
	ERRLOGwrap[n] = ERRLOGmark ;
	ERRLOGcount ++ ;

	i = 0 ;
	while(i<ERRLOGfatals) {
		if(ERRLOGfatal[i]==errno) break ;
		i ++ ;
	}

	if(i<ERRLOGfatals) {
		ERRLOGmark = ERRLOGcount ;
	}

	if(ERRLOGdbgmsk&(((unsigned long)0x1)<<ERRLOGlev[n])) ERRPRINT() ;

	if(i<ERRLOGfatals) {
		exit(errno) ;
	}

	va_end(ap) ;
}

void MESSAGE(char * funcname, char * message, ...) {
	va_list ap ;
	struct timeval tm ;
	register unsigned long n ;
	register unsigned long i ;
	register unsigned long j ;
	register unsigned long d ;
	unsigned long l ;
	char * s ;
	unsigned long u ;
	char c ;
	va_start(ap,message) ;
	gettimeofday(&tm,NULL) ;
	n = ERRLOGcount%ERRLOG_DEPTH ;
	ERRLOGent[n] = ERRLOGcount ;
	ERRLOGnum[n] = 0 ;
	ERRLOGsec[n] = tm.tv_sec ;
	ERRLOGmsc[n] = tm.tv_usec ;
	ERRLOGlev[n] = 31 ;
	i = 0 ;
	while((i<ERRLOG_NAME_LEN)&&(funcname[i]!='\0')) {
		ERRLOGname[n][i] = funcname[i] ;
		i ++ ;
	}
	ERRLOGname[n][i] = '\0' ;
	if(message!=NULL) {
		i = 0 ;
		j = 0 ;
		while((i<ERRLOG_MESS_LEN)&&(message[j]!='\0')) {
			while((i<ERRLOG_MESS_LEN)&&(message[j]!='\0')&&(message[j]!='%')) {
				ERRLOGmess[n][i] = message[j] ;
				i ++ ;
				j ++ ;
			}
			if(message[j]=='%') {
				j ++ ;
				if(message[j]=='\0') break ;
				l = 0 ;
				if( (message[j]>='0')&&(message[j]<='9') ) {
					j += _scaninteger(&l,message+j,10,10) ;
					if(l>(ERRLOG_MESS_LEN-i)) l = ERRLOG_MESS_LEN-i ;
				}
				else l = ERRLOG_MESS_LEN-i ;
				switch(message[j]) {
					case '%':
						if(i<ERRLOG_MESS_LEN) {
							ERRLOGmess[n][i] = '%' ;
							i ++ ;
						}
						break ;
					case 'c':
						c = (char) va_arg(ap, int) ;
						if(i<ERRLOG_MESS_LEN) {
							ERRLOGmess[n][i] = c ;
							i ++ ;
						}
						break ;
					case 's':
						s = (char *) va_arg(ap, char *) ;
						if(s==NULL) break ;
						d = 0 ;
						while( (i<ERRLOG_MESS_LEN)&&(d<l)&&(s[d]!='\0') ) {
							ERRLOGmess[n][i] = s[d] ;
							i ++ ;
							d ++ ;
						}
						break ;
					case 'u':
						u = (unsigned long) va_arg(ap, int) ;
						i += _printinteger(u,ERRLOGmess[n]+i,l,10) ;
						break ;
					case 'x':
						u = (unsigned long) va_arg(ap, int) ;
						i += _printinteger(u,ERRLOGmess[n]+i,l,16) ;
						break ;
					default:
						;
				}
				j ++ ;
			}
		}
		ERRLOGmess[n][i] = '\0' ;
	}
	else {
		ERRLOGmess[n][0] = '\0' ;
	}
	ERRLOGwrap[n] = ERRLOGmark ;
	ERRLOGcount ++ ;

	if(ERRLOGdbgmsk&(((unsigned long)0x1)<<ERRLOGlev[n])) ERRPRINT() ;

	va_end(ap) ;
}

void ERRPRINT() {
	register unsigned long p = 0 ;
	register unsigned long d ;
	char * err ;
	char buf[64] ; // 2+2+16+2+6+4+6+4 = 42 < 64
	if(ERRLOGcount==0) return ;
	p = (ERRLOGcount-1)%ERRLOG_DEPTH ;
	d = 0 ;
	d += _printinteger(ERRLOGlev[p],buf+d,2,10) ;
	buf[d] = '>' ; d ++ ;
	buf[d] = ' ' ; d ++ ;
	d += _printinteger(ERRLOGsec[p],buf+d,16,10) ;
	buf[d] = 's' ; d ++ ;
	buf[d] = '|' ; d ++ ;
	d += _printinteger(ERRLOGmsc[p],buf+d,6,10) ;
	buf[d] = 'u' ; d ++ ;
	buf[d] = 's' ; d ++ ;
	buf[d] = ' ' ; d ++ ;
	buf[d] = '[' ; d ++ ;
	d += _printinteger(ERRLOGent[p],buf+d,6,10) ;
	buf[d] = ']' ; d ++ ;
	buf[d] = ' ' ; d ++ ;
	buf[d] = ':' ; d ++ ;
	buf[d] = ' ' ; d ++ ;
	write(ERRLOGout,buf,d) ;
	write(ERRLOGout,ERRLOGname[p],strlen(ERRLOGname[p])) ;
	if(ERRLOGnum[p]!=0) {
		write(ERRLOGout," > ",3) ;
		err = strerror(ERRLOGnum[p]) ;
		if(err!=NULL) write(ERRLOGout,err,strlen(err)) ;
	}
	write(ERRLOGout," (",2) ;
	write(ERRLOGout,ERRLOGmess[p],strlen(ERRLOGmess[p])) ;
	write(ERRLOGout,")\n",2) ;
}

void ERRFLUSH() {
	register unsigned long p = 0 ;
	register unsigned long d ;
	char * err ;
	char buf[64] ; // 2+2+16+2+6+4+16+4 = 42 < 64
	if((ERRLOGcount==0)) return ;
	p = (ERRLOGwrap[(ERRLOGcount-1)%ERRLOG_DEPTH]-1)%ERRLOG_DEPTH ;
	while(p<ERRLOGcount%ERRLOG_DEPTH) {
		d = 0 ;
		d += _printinteger(ERRLOGlev[p],buf+d,2,10) ;
		buf[d] = '>' ; d ++ ;
		buf[d] = ' ' ; d ++ ;
		d += _printinteger(ERRLOGsec[p]+d,buf,16,10) ;
		buf[d] = 's' ; d ++ ;
		buf[d] = '|' ; d ++ ;
		d += _printinteger(ERRLOGmsc[p]+d,buf,6,10) ;
		buf[d] = 'u' ; d ++ ;
		buf[d] = 's' ; d ++ ;
		buf[d] = ' ' ; d ++ ;
		buf[d] = '[' ; d ++ ;
		d += _printinteger(ERRLOGent[p]+d,buf,6,10) ;
		buf[d] = ']' ; d ++ ;
		buf[d] = ' ' ; d ++ ;
		buf[d] = ':' ; d ++ ;
		buf[d] = ' ' ; d ++ ;
		write(ERRLOGout,buf,d) ;
		write(ERRLOGout,ERRLOGname[p],strlen(ERRLOGname[p])) ;
		if(ERRLOGnum[p]!=0) {
			write(ERRLOGout," > ",3) ;
			err = strerror(ERRLOGnum[p]) ;
			if(err!=NULL) write(ERRLOGout,err,strlen(err)) ;
		}
		write(ERRLOGout," (",2) ;
		write(ERRLOGout,ERRLOGmess[p],strlen(ERRLOGmess[p])) ;
		write(ERRLOGout,")\n",2) ;
	}
}

