
/*################################################################################################
 *##################                                                            ##################
 *##################                       Object Types                         ##################
 *##################                                                            ##################
 *##################                      Implementation                        ##################
 *##################                                                            ##################
 *################################################################################################
 */

/*
 * OKit:   Object Tree,  RPL Virtual Machine  and  compiler  shared library
 *
 * Copyright (C) 2000 Yann LANDRIN-SCHWEITZER
 *
 *
 * This library is free software;  you can redistribute it and/or modify it
 * under  the terms of the  GNU Lesser General Public License  as published
 * by the  Free Software Foundation;  either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This  library  is  distributed  in  the hope that it will be useful, but
 * WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty  of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library;  if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * Author:   Yann LANDRIN-SCHWEITZER
 * Contact:  landrin@multimania.com
 * Homepage: http://landrin.multimania.com/
 */



#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>

#include <Internal.h>
#include <OKit/ErrLog.h>

#include <OKit/ObjTypes.h>
#include <OKit/FileIOMod.h>



//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  File-descriptor Printing
/////

unsigned long ObjectPrint(Object obj, int fd) {
	if(obj==NULL) {
		errno = EINVAL ;
		DEBUG("ObjectPrint",1,"NULL Object") ;
		return 0 ;
	}
	switch(obj->type) {
		case OT_VOID:
			return VoidPrint(obj->obj,fd) ;
		case OT_INTG:
			return IntgPrint(obj->obj,fd) ;
		case OT_FLTP:
			return FltpPrint(obj->obj,fd) ;
		case OT_STRG:
			return StrgPrint(obj->obj,fd) ;
		case OT_LIST:
			return ListPrint(obj->obj,fd) ;
		case OT_CODE:
			return CodePrint(obj->obj,fd) ;
		case OT_BLTN:
			return BltnPrint(obj->obj,fd) ;
	}
	return 0 ;
}


unsigned long VoidPrint(Void vd, int fd) {
	if(write(fd,"",1)<1) {
		DEBUG("VoidPrint",1,NULL) ;
		return -1 ;
	}
	return 1 ;
}


unsigned long IntgPrint(Intg in, int fd) {
	char bf[16] ;
	int nb ;
	if(in==NULL) {
		errno = EINVAL ;
		DEBUG("IntgPrint",1,"NULL Intg") ;
		return -1 ;
	}
	if((*in)<0) {
		bf[0] = '-' ;
		nb = 1 + _printinteger(-(*in),bf+1,15,10) ;
	}
	else nb = _printinteger(*in,bf,16,10) ;
	if(write(fd,bf,nb)<nb) {
		DEBUG("IntgPrint",1,NULL) ;
		return -1 ;
	}
	return nb ;
}


unsigned long FltpPrint(Fltp fp, int fd) {
	char bf[32] ;
	int nb ;
	if(fp==NULL) {
		errno = EINVAL ;
		DEBUG("FltpPrint",1,"NULL Fltp") ;
		return -1 ;
	}
	nb = _printfloat(*fp,bf,32,24) ;
	if(write(fd,bf,nb)<nb) {
		DEBUG("FltpPrint",1,NULL) ;
		return -1 ;
	}
	return nb ;
}


unsigned long StrgPrint(Strg str, int fd) {
	if(str==NULL) {
		errno = EINVAL ;
		DEBUG("StrgPrint",1,"NULL Strg") ;
		return -1 ;
	}
	if(write(fd,"\"",1)<1) {
		DEBUG("StrgPrint",1,NULL) ;
		return -1 ;
	}
	if(write(fd,str->chr,str->len)<str->len) {
		DEBUG("StrgPrint",1,NULL) ;
		return -1 ;
	}
	if(write(fd,"\"",1)<1) {
		DEBUG("StrgPrint",1,NULL) ;
		return -1 ;
	}
	return str->len+2 ;
}


unsigned long ListPrint(List lst, int fd) {
	struct S_Node * nod ;
	register unsigned long i = 0 ;
	register unsigned long b = 0 ;
	if(lst==NULL) {
		errno = EINVAL ;
		DEBUG("ListPrint",1,"NULL List") ;
		return -1 ;
	}
	if(write(fd,"[ ",2)<2) {
		DEBUG("ListPrint",1,NULL) ;
		return -1 ;
	}
	i += 2 ;
	nod = lst->frst ;
	if(nod!=NULL) {
		b = ObjectPrint(nod->obj,fd) ;
		if(b<0) {
			DEBUG("ListPrint",1,NULL) ;
			return -1 ;
		}
		i += b ;
		if(write(fd," ",1)<1) {
			DEBUG("ListPrint",1,NULL) ;
			return -1 ;
		}
		i ++ ;
		while((nod->next!=NULL)&&(nod!=lst->last)) {
			nod = nod->next ;
			b = ObjectPrint(nod->obj,fd) ;
			if(b<0) {
				DEBUG("ListPrint",1,NULL) ;
				return -1 ;
			}
			i += b ;
			if(write(fd," ",1)<1) {
				DEBUG("ListPrint",1,NULL) ;
				return -1 ;
			}
			i ++ ;
		}
	}
	if(write(fd,"]",1)<1) {
		DEBUG("ListPrint",1,NULL) ;
		return -1 ;
	}
	i ++ ;
	return i ;
}


unsigned long CodePrint(Code cod, int fd) {
	unsigned long i = 0 ;
	if(cod==NULL) {
		errno = EINVAL ;
		DEBUG("CodePrint",1,"NULL Code") ;
		return -1 ;
	}
	if(write(fd,"<",1)<1) {
		DEBUG("CodePrint",1,NULL) ;
		return -1 ;
	}
	if(write(fd,cod->name,cod->nmcl)<cod->nmcl) {
		DEBUG("CodePrint",1,NULL) ;
		return -1 ;
	}
	if(cod->data!=NULL) {
		if(write(fd,":",1)<1) {
			DEBUG("CodePrint",1,NULL) ;
			return -1 ;
		}
		i = ObjectPrint(cod->data,fd) ;
		if(i<0) {
			DEBUG("CodePrint",1,NULL) ;
			return -1 ;
		}
		i ++ ;
	}	
	if(write(fd,">",1)<1) {
		DEBUG("CodePrint",1,NULL) ;
		return -1 ;
	}
	return cod->nmcl+i+2 ;
}


unsigned long BltnPrint(Bltn blt, int fd) {
	register unsigned long i = 0 ;
	if(blt==NULL) {
		errno = EINVAL ;
		DEBUG("BltnPrint",1,"NULL Bltn") ;
		return -1 ;
	}
	if(write(fd,"",1)<1) {
		DEBUG("BltnPrint",1,NULL) ;
		return -1 ;
	}
	if(blt->name!=NULL) {
		while(blt->name[i]!=0) i ++ ;
		if(write(fd,blt->name,i)<i) {
			DEBUG("BltnPrint",1,NULL) ;
			return -1 ;
		}
	}
	if(write(fd,"",1)<1) {
		DEBUG("BltnPrint",1,NULL) ;
		return -1 ;
	}
	return i+2 ;
}





//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Stack and NameSpace File-descriptor Printing
/////



unsigned long StackFPrint(Stack stk, int fd, char * head, char * rowfmt, char * tail, unsigned long rows) {
	register unsigned long i ;
	register unsigned long j ;
	char s[257] ;
	register unsigned long u ;
	register unsigned long r ;
	unsigned long l ;
	char fil = ' ' ;  //Fill char
	char b[257] ;

	if(stk==NULL) {
		errno = EINVAL ;
		DEBUG("StackFPrint",2,"NULL Stack") ;
		return -1 ;
	}
	//First line
	l = 0 ;
	while(l<256) {
		if(head[l]==0) break ;
		l ++ ;
	}
	if(write(fd,head,l)<l) {
		DEBUG("StackFPrint",2,NULL) ;
		return -1 ;
	}
	if(write(fd,"\n",1)<1) {
		DEBUG("StackFPrint",2,NULL) ;
		return -1 ;
	}
	i = rows ;
	// FPrinting empty stack lines
	while(i>stk->dpth) {
		// One empty line more...
		j = 0 ;
		while(rowfmt[j]!=0) {
			u = j ;
			while(rowfmt[j]!=0) {
				if(rowfmt[j]=='%') break ;
				j ++ ;
			}
			if(write(fd,rowfmt+u,j-u)<j-u) {
				DEBUG("StackFPrint",2,NULL) ;
				return -1 ;
			}
			if(rowfmt[j]==0) break ;
			j ++ ;
			l = 0 ;
			if(rowfmt[j]==0) break ;
			if( (rowfmt[j]<='0')||(rowfmt[j]>'9') ) {
				fil = rowfmt[j] ;
				j ++ ;
			}
			else fil = ' ' ;
			if(rowfmt[j]==0) break ;
			if( (rowfmt[j]>='0')&&(rowfmt[j]<='9') ) {
				j += _scaninteger(&l,rowfmt+j,10,10) ;
			}
			if(rowfmt[j]==0) break ;
			if(l>256) l = 256 ;
			switch(rowfmt[j]) {
				case 'n':
					if(l==0) l = 4 ;
					_printinteger(i,s,l,10) ;
					if(write(fd,s,l)<l) {
						DEBUG("StackFPrint",2,NULL) ;
						return -1 ;
					}
					break ;
				case 's':
					if(l==0) l = 20 ;
					u = l ;
					while(u>0) {
						u -- ;
						b[u] = fil ;
					}
					if(write(fd,b,l)<l) {
						DEBUG("StackFPrint",2,NULL) ;
						return -1 ;
					}
					break ;
				default:
					;
			}
			j ++ ;
		}
		if(write(fd,"\n",1)<1) {
			DEBUG("StackFPrint",2,NULL) ;
			return -1 ;
		}
		i -- ;
	}
	// FPrinting real stack lines
	while(i>0) {
		j = 0 ;
		while(rowfmt[j]!=0) {
			u = j ;
			while(rowfmt[j]!=0) {
				if(rowfmt[j]=='%') break ;
				j ++ ;
			}
			if(write(fd,rowfmt+u,j-u)<j-u) {
				DEBUG("StackFPrint",2,NULL) ;
				return -1 ;
			}
			if(rowfmt[j]==0) break ;
			j ++ ;
			l = 0 ;
			if(rowfmt[j]==0) break ;
			if( (rowfmt[j]<='0')||(rowfmt[j]>'9') ) {
				fil = rowfmt[j] ;
				j ++ ;
			}
			else fil = ' ' ;
			if(rowfmt[j]==0) break ;
			if( (rowfmt[j]>='0')&&(rowfmt[j]<='9') ) {
				j += _scaninteger(&l,rowfmt+j,10,10) ;
			}
			if(rowfmt[j]==0) break ;
			if(l>256) l = 256 ;
			switch(rowfmt[j]) {
				case 'n':
					if(l==0) l = 4 ;
					_printinteger(i,s,l,10) ;
					if(write(fd,s,l)<l) {
						DEBUG("StackFPrint",2,NULL) ;
						return -1 ;
					}
					break ;
				case 's':
					if(l==0) l = 20 ;
					r = ObjectOut(stk->data[stk->dpth-i],s,l) ;
					if(r==-1) {
						DEBUG("StackFPrint",2,NULL) ;
						return -1 ;
					}
					if(r<l) {
						u = l-r ;
						while(u>0) {
							u -- ;
							b[u] = fil ;
						}
						if(write(fd,b,l-r)<l-r) {
							DEBUG("StackFPrint",2,NULL) ;
							return -1 ;
						}
					}
					if(write(fd,s,r)<r) {
						DEBUG("StackFPrint",2,NULL) ;
						return -1 ;
					}
					break ;
				default:
					;
			}
			j ++ ;
		}
		if(write(fd,"\n",1)<1) {
			DEBUG("StackFPrint",2,NULL) ;
			return -1 ;
		}
		i -- ;
	}
	l = 0 ;
	while(l<256) {
		if(tail[l]==0) break ;
		l ++ ;
	}
	if(write(fd,tail,l)<l) {
		DEBUG("StackFPrint",2,NULL) ;
		return -1 ;
	}
	if(write(fd,"\n",1)<1) {
		DEBUG("StackFPrint",2,NULL) ;
		return -1 ;
	}
	// FIXME this is temporary, should compute real length -- what for ? --
	// Should be rows+2 lines of cols+1 bytes (final CR included)
	return (rows+2) * (l+1) ;
}


unsigned long NameSpaceFPrint(NameSpace nms, int fd, char * head, char * rowfmt, char * tail) {
	register unsigned long i ;
	register unsigned long j ;
	char s[257] ;
	register unsigned long u ;
	register unsigned long r ;
	unsigned long l ;
	char fil = ' ' ;  //Fill char
	char b[257] ;

	if(nms==NULL) {
		errno = EINVAL ;
		DEBUG("NameSpaceFPrint",2,"NULL NameSpace") ;
		return -1 ;
	}
	//First line
	l = 0 ;
	while(l<256) {
		if(head[l]==0) break ;
		l ++ ;
	}
	if(write(fd,head,l)<l) {
		DEBUG("NameSpaceFPrint",2,NULL) ;
		return -1 ;
	}
	if(write(fd,"\n",1)<1) {
		DEBUG("NameSpaceFPrint",2,NULL) ;
		return -1 ;
	}
	i = 0 ;
	// Printing variables
	while(i<nms->dpth) {
		// One var line more...
		j = 0 ;
		while(rowfmt[j]!=0) {
			u = j ;
			while(rowfmt[j]!=0) {
				if(rowfmt[j]=='%') break ;
				j ++ ;
			}
			if(write(fd,rowfmt+u,j-u)<j-u) {
				DEBUG("NameSpaceFPrint",2,NULL) ;
				return -1 ;
			}
			if(rowfmt[j]==0) break ;
			j ++ ;
			l = 0 ;
			if(rowfmt[j]==0) break ;
			if( (rowfmt[j]<='0')||(rowfmt[j]>'9') ) {
				fil = rowfmt[j] ;
				j ++ ;
			}
			else fil = ' ' ;
			if(rowfmt[j]==0) break ;
			if( (rowfmt[j]>='0')&&(rowfmt[j]<='9') ) {
				j += _scaninteger(&l,rowfmt+j,10,10) ;
			}
			if(rowfmt[j]==0) break ;
			if(l>256) l = 256 ;
			switch(rowfmt[j]) {
				case 'n':
					if(l==0) l = 4 ;
					r = 0 ;
					while(r<l) {
						if(nms->names[i][r]==0) break ;
						r ++ ;
					}
					u = l-r ;
					while(u>0) {
						u -- ;
						b[u] = fil ;
					}
					if(write(fd,b,l-r)<l-r) {
						DEBUG("NameSpaceFPrint",2,NULL) ;
						return -1 ;
					}
					if(write(fd,s,r)<r) {
						DEBUG("NameSpaceFPrint",2,NULL) ;
						return -1 ;
					}
					break ;
				case 's':
					if(l==0) l = 20 ;
					r = ObjectOut(nms->conts[i],s,l) ;
					if(r==-1) {
						DEBUG("NameSpaceFPrint",2,NULL) ;
						return -1 ;
					}
					if(r<l) {
						u = l-r ;
						while(u>0) {
							u -- ;
							b[u] = fil ;
						}
						if(write(fd,b,l-r)<l-r) {
							DEBUG("NameSpaceFPrint",2,NULL) ;
							return -1 ;
						}
					}
					if(write(fd,s,r)<r) {
						DEBUG("NameSpaceFPrint",2,NULL) ;
						return -1 ;
					}
					break ;
				default:
					;
			}
			j ++ ;
		}
		if(write(fd,"\n",1)<1) {
			DEBUG("NameSpaceFPrint",2,NULL) ;
			return -1 ;
		}
		i -- ;
	}
	l = 0 ;
	while(l<256) {
		if(tail[l]==0) break ;
		l ++ ;
	}
	if(write(fd,tail,l)<l) {
		DEBUG("NameSpaceFPrint",2,NULL) ;
		return -1 ;
	}
	if(write(fd,"\n",1)<1) {
		DEBUG("NameSpaceFPrint",2,NULL) ;
		return -1 ;
	}
	// FIXME this is temporary, should compute real length -- what for ? --
	return (nms->dpth+2) * (l+1) ;
}


