
/*################################################################################################
 *##################                                                            ##################
 *##################                     INTERNAL TOOLS                         ##################
 *##################                                                            ##################
 *##################                                                            ##################
 *##################                     Implementation                         ##################
 *##################                                                            ##################
 *################################################################################################
 */

/*
 * OKit:   Object Tree,  RPL Virtual Machine  and  compiler  shared library
 *
 * Copyright (C) 2000 Yann LANDRIN-SCHWEITZER
 *
 *
 * This library is free software;  you can redistribute it and/or modify it
 * under  the terms of the  GNU Lesser General Public License  as published
 * by the  Free Software Foundation;  either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This  library  is  distributed  in  the hope that it will be useful, but
 * WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty  of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library;  if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * Author:   Yann LANDRIN-SCHWEITZER
 * Contact:  landrin@multimania.com
 * Homepage: http://landrin.multimania.com/
 */


#include <stdlib.h>
#include <unistd.h>
#include <Internal.h>


//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Printing and scanning
/////


unsigned long _printinteger(unsigned long u, char * s, unsigned long m, unsigned long b) {
	register unsigned long i = 0 ;
	register unsigned long d = 0 ;
	if(b==0) return 0 ;
	if(b>32) return 0 ;
	while( (b>>(d+1)) != 0 ) d ++ ;
	d = (8*sizeof(unsigned long))/d ;
	if(s==NULL) return d ;
	m = (d>m)?m:d ;
	i = m ;
	while(i>0) {
		i -- ;
		s[i] = (u%b<10)?(u%b+'0'):(u%b-10+'A') ;
		u /= b ;
	}
	return m ;
}


unsigned long _scaninteger(unsigned long * u, char * s, unsigned long m, unsigned long b) {
	register unsigned long i = 0 ;
	char c ;
	if(u==NULL) return 0 ;
	if(s==NULL) return 0 ;
	if(b==0) return 0 ;
	if(b>32) return 0 ;
	(*u) = 0 ;
	while(i<m) {
		if(s[i]=='\0') break ;
		if( (s[i]>='0') && (s[i]<='9') ) c = s[i]-'0' ;
		else if( (s[i]>='A') && (s[i]<='Z') ) c = s[i]-'A'+10 ;
		else if( (s[i]>='a') && (s[i]<='z') ) c = s[i]-'a'+10 ;
		else break ;
		if(c>=b) break ;
		(*u) *= b ;
		(*u) += c ;
		i ++ ;
	}
	return i ;
}


int snprintf(char * str, size_t size, const char * format, ...) ;
int sscanf(const char * str, const char * format, ...) ;

unsigned long _printfloat(long double u, char * s, unsigned long m, unsigned long e) {
	register unsigned long i = 0 ;
	if(s==NULL) m = 0 ;
	i = snprintf(s,m,"%*.*Le",m-1,e,u) ;
	return (m>i)?i:m ;
}


unsigned long _scanfloat(long double * u, char * s, unsigned long m) {
	register unsigned long i = 0 ;
	i = sscanf(s,"%Le",u) ;
	return (m>i)?i:m ;
}



//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Search and key lookup
/////


signed char _keyncmp(char * ky1, char * ky2, unsigned long len) {
	register unsigned long i = 0 ;
	while(i<len) {
		if(ky1[i]<ky2[i]) return -1 ;
		if(ky1[i]>ky2[i]) return +1 ;
		if(ky1[i]==0) break ;
		i ++ ;
	}
	return 0 ;
}


unsigned long _keysearch(char * key, char * * keys, unsigned long keynum, unsigned long keylen, unsigned long * pos) {
	register unsigned long min, max, med ;
	if(keynum==0) {
		if(pos!=NULL) *pos = 0 ;
		return -1 ;
	}
	min = 0 ;
	max = keynum ;
	switch(_keyncmp(key,keys[min],keylen)) {
		case -1:
			if(pos!=NULL) *pos = min ;
			return -1 ;
		case +1:
			break ;
		case 0:
			if(pos!=NULL) *pos = min ;
			return min ;
		default:
			return -1 ;
	}
	switch(_keyncmp(key,keys[max-1],keylen)) {
		case -1:
			break ;
		case +1:
			if(pos!=NULL) *pos = max ;	
			return -1 ;
		case 0:
			if(pos!=NULL) *pos = max-1 ;
			return max-1 ;
		default:
			return -1 ;
	}
	while(min<max) {
		med = (min+max-1)>>1 ;
		switch(_keyncmp(key,keys[med],keylen)) {
			case -1:
				max = med ;
				break ;
			case +1:
				min = med+1 ;
				break ;
			case 0:
				if(pos!=NULL) *pos = med ;
				return med ;
			default:
				return -1 ;
		}
	}
	if(pos!=NULL) *pos = min ;                      // min == max here
	if(_keyncmp(key,keys[min],keylen)==0) return min ;// ensure we haven't some pathological
	                                                //case such as keynum == 1
	return -1 ;
}


