
/*################################################################################################
 *##################                                                            ##################
 *##################                       Object Types                         ##################
 *##################                                                            ##################
 *##################                      Implementation                        ##################
 *##################                                                            ##################
 *################################################################################################
 */

/*
 * OKit:   Object Tree,  RPL Virtual Machine  and  compiler  shared library
 *
 * Copyright (C) 2000 Yann LANDRIN-SCHWEITZER
 *
 *
 * This library is free software;  you can redistribute it and/or modify it
 * under  the terms of the  GNU Lesser General Public License  as published
 * by the  Free Software Foundation;  either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This  library  is  distributed  in  the hope that it will be useful, but
 * WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty  of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library;  if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * Author:   Yann LANDRIN-SCHWEITZER
 * Contact:  landrin@multimania.com
 * Homepage: http://landrin.multimania.com/
 */



#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>

#include <Internal.h>
#include <OKit/ErrLog.h>

#include <OKit/ObjTypes.h>



//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Generic Object methods
/////


Object ObjectAlloc() {
	Object obj ;
	obj = (Object) malloc(sizeof(struct S_Object)) ;
	if(obj==NULL) {
		DEBUG("ObjectAlloc",0,NULL) ;
		return NULL ;
	}
	obj->type = 0 ;
	obj->refc = 0 ;
	obj->flag = 0 ;
	obj->size = 0 ;
	obj->obj = NULL ;
	return obj ;
}


Object ObjectCopy(Object obj) {
	Object objl ;
	if(obj==NULL) {
		errno = EINVAL ;
		DEBUG("ObjectCopy",1,"NULL Object") ;
		return NULL ;
	}
	objl = (Object) malloc(sizeof(struct S_Object)) ;
	if(objl==NULL) {
		DEBUG("ObjectCopy",0,NULL) ;
		return NULL ;
	}
	objl->type = obj->type ;
	objl->refc = 0 ;
	objl->flag = 0 ;
	objl->size = obj->size ;
	if(obj->obj==NULL) {
		objl->obj = NULL ;
	}
	else {
		switch(obj->type) {
			case OT_VOID:
				objl->obj = NULL ;
				return objl ; //Already OK, no check
			case OT_INTG:
				objl->obj = IntgCopy(obj->obj) ;
				break ;
			case OT_FLTP:
				objl->obj = FltpCopy(obj->obj) ;
				break ;
			case OT_STRG:
				objl->obj = StrgCopy(obj->obj) ;
				break ;
			case OT_LIST:
				objl->obj = ListCopy(obj->obj) ;
				break ;
			case OT_CODE:
				objl->obj = CodeCopy(obj->obj) ;
				break ;
			case OT_BLTN:
				objl->obj = obj->obj ;
				break ;
			default:
				objl->obj = NULL ;
				return objl ;
		}
		if(objl->obj==NULL) {
			free(objl) ;
			DEBUG("ObjectCopy",1,NULL) ;
			return NULL ;
		}
	}
	return objl ;
}


Object ObjectClone(Object obj) {
	Object objl ;
	if(obj==NULL) {
		errno = EINVAL ;
		DEBUG("ObjectClone",1,"NULL Object") ;
		return NULL ;
	}
	objl = (Object) malloc(sizeof(struct S_Object)) ;
	if(objl==NULL) {
		DEBUG("ObjectClone",0,NULL) ;
		return NULL ;
	}
	objl->type = obj->type ;
	objl->refc = 0 ;
	objl->flag = 0 ;
	objl->size = obj->size ;
	if(obj->obj==NULL) {
		objl->obj = NULL ;
	}
	else {
		switch(obj->type) {
			case OT_VOID:
				objl->obj = NULL ;
				return objl ; //Already OK, no check
			case OT_INTG:
				objl->obj = IntgClone(obj->obj) ;
				break ;
			case OT_FLTP:
				objl->obj = FltpClone(obj->obj) ;
				break ;
			case OT_STRG:
				objl->obj = StrgClone(obj->obj) ;
				break ;
			case OT_LIST:
				objl->obj = ListClone(obj->obj) ;
				break ;
			case OT_CODE:
				objl->obj = CodeClone(obj->obj) ;
				break ;
			case OT_BLTN:
				objl->obj = obj->obj ;
				break ;
			default:
				objl->obj = NULL ;
				return objl ;
		}
		if(objl->obj==NULL) {
			free(objl) ;
			DEBUG("ObjectClone",1,NULL) ;
			return NULL ;
		}
	}
	return objl ;
}


Object ObjectDup(Object obj) {
	if(obj==NULL) return NULL ;
	obj->refc ++ ;
	if(obj->obj!=NULL) {
		switch(obj->type) {
			case OT_VOID:
				break ;
			case OT_INTG:
				//IntgDup(obj->obj) ;
				break ;
			case OT_FLTP:
				//FltpDup(obj->obj) ;
				break ;
			case OT_STRG:
				//StrgDup(obj->obj) ;
				break ;
			case OT_LIST:
				ListDup(obj->obj) ;
				break ;
			case OT_CODE:
				CodeDup(obj->obj) ;
				break ;
			case OT_BLTN:
				//BltnDup(obj->obj) ;
				break ;
			default:
				;
		}
	}
	return obj ;
}


Object ObjectDrop(Object obj) {
	if(obj==NULL) return NULL ;
	if(obj->refc!=0) obj->refc -- ;
	else {
		errno = EFAULT ;
		DEBUG("ObjectDrop",0,"Reference fall-through-the-floor") ;
		return NULL ;
	}
	if(obj->obj!=NULL) {
		switch(obj->type) {
			case OT_VOID:
				break ;
			case OT_INTG:
				//IntgDrop(obj->obj) ;
				break ;
			case OT_FLTP:
				//FltpDrop(obj->obj) ;
				break ;
			case OT_STRG:
				//StrgDrop(obj->obj) ;
				break ;
			case OT_LIST:
				ListDrop(obj->obj) ;
				break ;
			case OT_CODE:
				CodeDrop(obj->obj) ;
				break ;
			case OT_BLTN:
				//BltnDrop(obj->obj) ;
				break ;
			default:
				;
		}
	}
	return obj ;
}


void ObjectFree(Object obj) {
	if(obj==NULL) return ;
	if(obj->refc==0) {
		if(obj->obj!=NULL) switch(obj->type) {
			case OT_VOID:
				//VoidFree(obj->obj) ;
				break ;
			case OT_INTG:
				//IntgFree(obj->obj) ;
				free(obj->obj) ;
				break ;
			case OT_FLTP:
				//FltpFree(obj->obj) ;
				free(obj->obj) ;
				break ;
			case OT_STRG:
				StrgFree(obj->obj) ;
				break ;
			case OT_LIST:
				ListFree(obj->obj) ;
				break ;
			case OT_CODE:
				CodeFree(obj->obj) ;
				break ;
			case OT_BLTN:
				//BltnFree(obj->obj) ;
				break ;
			default:
				;
				free(obj->obj) ;
		}
		obj->type = 0 ;
		obj->flag = 0 ;
		obj->size = 0 ;
		obj->obj = NULL ;
		free(obj) ;
	}
}


Type ObjectType(Object obj) {
	if(obj==NULL) {
		errno = EINVAL ;
		DEBUG("ObjectType",1,"NULL Object") ;
		return -1 ;
	}
	return obj->type ;
}


char * ObjectTypeName(Type typ) {
	switch(typ) {
		case OT_VOID:
			return "Void" ;
		case OT_INTG:
			return "Intg" ;
		case OT_FLTP:
			return "Fltp" ;
		case OT_STRG:
			return "Strg" ;
		case OT_LIST:
			return "List" ;
		case OT_CODE:
			return "Code" ;
		case OT_BLTN:
			return "Bltn" ;
		default:
			return "" ;
	}
}


void * ObjectContent(Object obj) {
	if(obj==NULL) {
		errno = EINVAL ;
		DEBUG("ObjectContent",1,"NULL Object") ;
		return NULL ;
	}
	return obj->obj ;
}


unsigned long ObjectStat(Object obj) {
	printf("Object [%p]\n",obj) ;
	if(obj==NULL) {
		printf("    Object type ID : N/A\n") ;
		printf("    Reference count: N/A\n") ;
		printf("    Internal flags : N/A\n") ;
		printf("    Object size    : N/A\n") ;
		printf("    Content address: N/A\n") ;
		return -1 ;
	}
	printf("    Object type ID : %x (%s)\n",obj->type,ObjectTypeName(obj->type)) ;
	printf("    Reference count: %lu\n",obj->refc) ;
	printf("    Internal flags : %lx\n",obj->flag) ;
	printf("    Object size    : %lu\n",obj->size) ;
	printf("    Content address: %p\n",obj->obj) ;
	return 0 ;
}


unsigned long ObjectOut(Object obj, char * out, unsigned long max) {
	if(obj==NULL) {
		errno = EINVAL ;
		DEBUG("ObjectOut",1,"NULL Object") ;
		return 0 ;
	}
	switch(obj->type) {
		case OT_VOID:
			return VoidOut(obj->obj,out,max) ;
		case OT_INTG:
			return IntgOut(obj->obj,out,max) ;
		case OT_FLTP:
			return FltpOut(obj->obj,out,max) ;
		case OT_STRG:
			return StrgOut(obj->obj,out,max) ;
		case OT_LIST:
			return ListOut(obj->obj,out,max) ;
		case OT_CODE:
			return CodeOut(obj->obj,out,max) ;
		case OT_BLTN:
			return BltnOut(obj->obj,out,max) ;
	}
	out[0] = 0 ;
	return 0 ;
}


unsigned char ObjectEqual(Object obj, Object objw) {
	if((obj==NULL)||(objw==NULL)) {
		errno = EINVAL ;
		DEBUG("ObjectEqual",1,"NULL Object") ;
		return -1 ;
	}
	if(obj->type!=objw->type) return 0 ;
	switch(obj->type) {
		case OT_VOID:
			//return VoidEqual(obj->obj,objw->obj) ;
			return 1 ;
		case OT_INTG:
			return IntgEqual(obj->obj,objw->obj) ;
		case OT_FLTP:
			return FltpEqual(obj->obj,objw->obj) ;
		case OT_STRG:
			return StrgEqual(obj->obj,objw->obj) ;
		case OT_LIST:
			return ListEqual(obj->obj,objw->obj) ;
		case OT_CODE:
			return CodeEqual(obj->obj,objw->obj) ;
		case OT_BLTN:
			return BltnEqual(obj->obj,objw->obj) ;
	}
	return 0 ;
}


Strg ObjectStrg(Object obj) {
	register unsigned long s = 0 ;
	Strg str ;
	str = (Strg) malloc(sizeof(struct S_Strg)) ;
	if(str==NULL) {
		DEBUG("ObjectStrg",0,NULL) ;
		return NULL ;
	}
	s = ObjectOut(obj,NULL,0) ;
	str->chr = (char *) malloc(s+1) ;
	if(str->chr==NULL) {
		free(str) ;
		DEBUG("ObjectStrg",0,NULL) ;
		return NULL ;
	}
	str->chr[s] = 0 ;
	if(ObjectOut(obj,str->chr,s)!=s) {
		DEBUG("ObjectStrg",1,NULL) ;
		free(str->chr) ;
		free(str) ;
		return NULL ;
	}
	str->len = s ;
	return str ;
}




//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Void object type
/////

// VOID Object Type
// typedef void Void ;
// #define ST_VOID         0



Void VoidNew() {
	return NULL ;
}


Object VoidWrap(Void vd) {
	Object obj ;
	obj = (Object) malloc(sizeof(struct S_Object)) ;
	if(obj==NULL) {
		DEBUG("VoidWrap",0,NULL) ;
		return NULL ;
	}
	obj->type = OT_VOID ;
	obj->refc = 0 ;
	obj->flag = 0 ;
	obj->size = ST_VOID ;
	obj->obj = vd ;
	return obj ;
}


unsigned long VoidOut(Void vd, char * out, unsigned long max) {
	if(out==NULL) {
		return 1 ;
	}
	if(max==0) {
		out[0] = 0 ;
		return 0 ;
	}
	out[0] = '' ;
	out[1] = 0 ;
	return 1 ;
}



//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Intg object type
/////

// INTG Object Type
// typedef long * Intg ;
// #define ST_INTG         sizeof(long)



Intg IntgAlloc() {
	Intg in ;
	in = (Intg) malloc(ST_INTG) ;
	if(in==NULL) {
		DEBUG("IntgAlloc",0,NULL) ;
		return NULL ;
	}
	*in = 0 ;
	return in ;
}


Intg IntgCopy(Intg in) {
	Intg inl ;
	if(in==NULL) {
		errno = EINVAL ;
		DEBUG("IntgCopy",1,NULL) ;
		return NULL ;
	}
	inl = (Intg) malloc(ST_INTG) ;
	if(inl==NULL) {
		DEBUG("IntgCopy",0,NULL) ;
		return NULL ;
	}
	*inl = *in ;
	return inl ;
}


Intg IntgClone(Intg in) {
	Intg inl ;
	if(in==NULL) {
		errno = EINVAL ;
		DEBUG("IntgClone",1,NULL) ;
		return NULL ;
	}
	inl = (Intg) malloc(ST_INTG) ;
	if(inl==NULL) {
		DEBUG("IntgClone",0,NULL) ;
		return NULL ;
	}
	*inl = *in ;
	return inl ;
}


Intg IntgNew(long nb) {
	Intg in = NULL ;
	in = malloc(ST_INTG) ;
	if(in==NULL) {
		DEBUG("IntgNew",0,NULL) ;
		return NULL ;
	}
	*((long *)in) = nb ;
	return in ;
}


Object IntgWrap(Intg in) {
	Object obj ;
	if(in==NULL) {
		errno = EINVAL ;
		DEBUG("IntgWrap",1,"NULL Intg") ;
		return NULL ;
	}
	obj = (Object) malloc(sizeof(struct S_Object)) ;
	if(obj==NULL) {
		DEBUG("IntgWrap",0,NULL) ;
		return NULL ;
	}
	obj->type = OT_INTG ;
	obj->refc = 0 ;
	obj->flag = 0 ;
	obj->size = ST_INTG ;
	obj->obj = (void *) in ;
	return obj ;
}


unsigned long IntgOut(Intg in, char * out, unsigned long max) {
	int nb ;
	if(in==NULL) {
		errno = EINVAL ;
		DEBUG("IntgOut",1,"NULL Intg") ;
		return -1 ;
	}
	if(out==NULL) {
		return _printinteger(*in,NULL,0,10) ;
	}
	if(max==0) {
		out[0] = 0 ;
		return 0 ;
	}
	if((*in)<0) {
		out[0] = '-' ;
		nb = 1 + _printinteger(-(*in),out+1,max-1,10) ;
	}
	else nb = _printinteger(*in,out,max,10) ;
	out[nb] = 0 ;
	return nb ;
}


unsigned char IntgEqual(Intg in, Intg inw) {
	if((in==NULL)||(inw==NULL)) {
		errno = EINVAL ;
		DEBUG("IntgEqual",1,"NULL Intg") ;
		return -1 ;
	}
	return (*in==*inw)?1:0 ;
}



//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Fltp object type
/////

// FLTP Object Type
// typedef double Fltp ;
// #define ST_FLTP         sizeof(double)



Fltp FltpAlloc() {
	Fltp fp ;
	fp = (Fltp) malloc(ST_FLTP) ;
	if(fp==NULL) {
		DEBUG("FltpAlloc",0,NULL) ;
		return NULL ;
	}
	*fp = 0 ;
	return fp ;
}


Fltp FltpCopy(Fltp fp) {
	Fltp fpl ;
	if(fp==NULL) {
		errno = EINVAL ;
		DEBUG("FltpCopy",1,NULL) ;
		return NULL ;
	}
	fpl = (Fltp) malloc(ST_FLTP) ;
	if(fpl==NULL) {
		DEBUG("FltpCopy",0,NULL) ;
		return NULL ;
	}
	*fpl = *fp ;
	return fpl ;
}


Fltp FltpClone(Fltp fp) {
	Fltp fpl ;
	if(fp==NULL) {
		errno = EINVAL ;
		DEBUG("FltpClone",1,NULL) ;
		return NULL ;
	}
	fpl = (Fltp) malloc(ST_FLTP) ;
	if(fpl==NULL) {
		DEBUG("FltpClone",0,NULL) ;
		return NULL ;
	}
	*fpl = *fp ;
	return fpl ;
}


Fltp FltpNew(double nb) {
	Fltp fp = NULL ;
	fp = malloc(ST_FLTP) ;
	if(fp==NULL) {
		DEBUG("FltpNew",0,NULL) ;
		return NULL ;
	}
	*((double *)fp) = nb ;
	return fp ;
}


Object FltpWrap(Fltp fp) {
	Object obj ;
	if(fp==NULL) {
		errno = EINVAL ;
		DEBUG("FltpWrap",1,"NULL Fltp") ;
		return NULL ;
	}
	obj = (Object) malloc(sizeof(struct S_Object)) ;
	if(obj==NULL) {
		DEBUG("FltpWrap",0,NULL) ;
		return NULL ;
	}
	obj->type = OT_FLTP ;
	obj->refc = 0 ;
	obj->flag = 0 ;
	obj->size = ST_FLTP ;
	obj->obj = (void *) fp ;
	return obj ;
}


unsigned long FltpOut(Fltp fp, char * out, unsigned long max) {
	int nb ;
	if(fp==NULL) {
		errno = EINVAL ;
		DEBUG("FltpOut",1,"NULL Fltp") ;
		return -1 ;
	}
	if(out==NULL) {
		return _printfloat(*fp,NULL,0,0) ;
	}
	nb = (max>32)?32:max ;
	nb = _printfloat(*fp,out,nb,nb-8) ;
	out[nb] = 0 ;
	return nb ;
}


unsigned char FltpEqual(Fltp fp, Fltp fpw) {
	if((fp==NULL)||(fpw==NULL)) {
		errno = EINVAL ;
		DEBUG("FltpEqual",1,"NULL Fltp") ;
		return -1 ;
	}
	return (*fp==*fpw)?1:0 ;
}



//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Strg manipulation
/////

// STRG Object Type
// struct S_Strg {
//	unsigned long len ;
//	char * chr ;
// } ;
// typedef struct S_Strg Strg ;
// #define ST_STRG         sizeof(struct S_Strg)


Strg StrgAlloc(unsigned long size) {
	register unsigned long i = 0 ;
	Strg str ;
	str = (Strg) malloc(sizeof(struct S_Strg)) ;
	if(str==NULL) {
		DEBUG("StrgAlloc",0,NULL) ;
		return NULL ;
	}
	str->chr = (char *) malloc(size+1) ;
	if(str->chr==NULL) {
		free(str) ;
		DEBUG("StrgAlloc",0,NULL) ;
		return NULL ;
	}
	while(i<size) {
		*(str->chr+i) = 0 ;
		i ++ ;
	}
	*(str->chr+i) = 0 ;
	str->len = size ;
	return str ;
}


Strg StrgCopy(Strg str) {
	register unsigned long i = 0 ;
	Strg strl ;
	if(str==NULL) {
		errno = EINVAL ;
		DEBUG("StrgCopy",1,"NULL Strg") ;
		return NULL ;
	}
	strl = (Strg) malloc(sizeof(struct S_Strg)) ;
	if(strl==NULL) {
		DEBUG("StrgCopy",0,NULL) ;
		return NULL ;
	}
	strl->chr = (char *) malloc(str->len+1) ;
	if(strl->chr==NULL) {
		free(strl) ;
		DEBUG("StrgCopy",0,NULL) ;
		return NULL ;
	}
	while(i<str->len) {
		*(strl->chr+i) = *(str->chr+i) ;
		i ++ ;
	}
	*(strl->chr+i) = 0 ;
	strl->len = str->len ;
	return strl ;
}


Strg StrgClone(Strg str) {
	register unsigned long i = 0 ;
	Strg strl ;
	if(str==NULL) {
		errno = EINVAL ;
		DEBUG("StrgClone",1,"NULL Strg") ;
		return NULL ;
	}
	strl = (Strg) malloc(sizeof(struct S_Strg)) ;
	if(strl==NULL) {
		DEBUG("StrgClone",0,NULL) ;
		return NULL ;
	}
	strl->chr = (char *) malloc(str->len+1) ;
	if(strl->chr==NULL) {
		free(strl) ;
		DEBUG("StrgClone",0,NULL) ;
		return NULL ;
	}
	while(i<str->len) {
		*(strl->chr+i) = *(str->chr+i) ;
		i ++ ;
	}
	*(strl->chr+i) = 0 ;
	strl->len = str->len ;
	return strl ;
}


void StrgFree(Strg str) {
	if(str!=NULL) {
		free(str->chr) ;
	}
	free(str) ;
}


Strg StrgNew(char * chs) {
	register unsigned long n = 0 ;
	register unsigned long i = 0 ;
	Strg str = NULL ;
	if(chs==NULL) {
		errno = EINVAL ;
		DEBUG("StrgNew",1,"NULL char * pointer") ;
		return NULL ;
	}
	str = malloc(sizeof(struct S_Strg)) ;
	if(str==NULL) {
		DEBUG("StrgNew",0,NULL) ;
		return NULL ;
	}
	while(n<~(long)0) {
		if(chs[n]==0) break ;
		n ++ ;
	}
	str->chr = (char *) malloc(n+1) ;
	if(str->chr==NULL) {
		free(str) ;
		DEBUG("StrgNew",0,NULL) ;
		return NULL ;
	}
	while(i<n) {
		str->chr[i] = chs[i] ;
		i ++ ;
	}
	str->chr[i] = 0 ;
	str->len = n ;
	return str ;
}


Object StrgWrap(Strg str) {
	Object obj ;
	if(str==NULL) {
		errno = EINVAL ;
		DEBUG("StrgWrap",1,"NULL Strg") ;
		return NULL ;
	}
	obj = (Object) malloc(sizeof(struct S_Object)) ;
	if(obj==NULL) {
		DEBUG("StrgWrap",0,NULL) ;
		return NULL ;
	}
	obj->type = OT_STRG ;
	obj->refc = 0 ;
	obj->flag = 0 ;
	obj->size = ST_STRG ;
	obj->obj = (void *) str ;
	return obj ;
}


unsigned long StrgLen(Strg str) {
	if(str==NULL) {
		errno = EINVAL ;
		DEBUG("StrgLen",1,"NULL Strg") ;
		return 0 ;
	}
	return str->len ;
}


Strg StrgCat(Strg str1, Strg str2) {
	register unsigned long i = 0 ;
	register unsigned long j ;
	Strg str ;
	if((str1==NULL)||(str2==NULL)) {
		errno = EINVAL ;
		DEBUG("StrgCat",1,"NULL Strg") ;
		return NULL ;
	}
	str = (Strg) malloc(sizeof(struct S_Strg)) ;
	if(str==NULL) {
		DEBUG("StrgCat",0,NULL) ;
		return NULL ;
	}
	str->chr = (char *) malloc(str1->len+str2->len+1) ;
	if(str->chr==NULL) {
		free(str) ;
		DEBUG("StrgCat",0,NULL) ;
		return NULL ;
	}
	j = 0 ;
	while(j<str1->len) {
		*(str->chr+i) = *(str1->chr+j) ;
		i ++ ;
		j ++ ;
	}
	j = 0 ;
	while(j<str2->len) {
		*(str->chr+i) = *(str2->chr+j) ;
		i ++ ;
		j ++ ;
	}
	*(str->chr+i) = 0 ;
	str->len = str1->len+str2->len ;
	return str ;
}


Strg StrgCut(Strg str, unsigned long beg, unsigned long end) {
	register unsigned long i = 0 ;
	Strg strl ;
	if(str==NULL) {
		errno = EINVAL ;
		DEBUG("StrgCut",1,"NULL Strg") ;
		return NULL ;
	}
	if( (beg>=str->len)||(end>=str->len) ) {
		errno = EINVAL ;
		DEBUG("StrgCut",1,"Invalid index") ;
		return NULL ;
	}
	strl = (Strg) malloc(sizeof(struct S_Strg)) ;
	if(strl==NULL) {
		DEBUG("StrgCut",0,NULL) ;
		return NULL ;
	}
	strl->chr = (char *) malloc(end-beg+1) ;
	if(strl->chr==NULL) {
		free(str) ;
		DEBUG("StrgCut",0,NULL) ;
		return NULL ;
	}
	while(i<end-beg) {
		*(strl->chr+i) = *(str->chr+i+beg) ;
		i ++ ;
	}
	*(strl->chr+i) = 0 ;
	strl->len = end-beg ;
	return strl ;
}

unsigned long StrgOut(Strg str, char * out, unsigned long max) {
	register unsigned long i = 0 ;
	if(str==NULL) {
		errno = EINVAL ;
		DEBUG("StrgOut",1,"NULL Strg") ;
		return -1 ;
	}
	if(out==NULL) {
		return str->len+2 ;
	}
	if(max==0) { out[0] = 0 ; return 0 ; }
	out[0] = '\"' ;
	max -- ;
	while((i<str->len)&&(i<max)) {
		out[i+1] = str->chr[i] ;
		i ++ ;
	}
	if(i<max) { i ++ ; out[i] = '\"' ; }
	i ++ ;
	out[i] = 0 ;
	return i ;
}


unsigned char StrgEqual(Strg str, Strg strw) {
	register unsigned long i = 0 ;
	if((str==NULL)||(strw==NULL)) {
		errno = EINVAL ;
		DEBUG("StrgEqual",1,"NULL Strg") ;
		return -1 ;
	}
	if(str->len!=strw->len) return 0 ;
	if(str->chr==strw->chr) return 1 ;
	while(i<str->len) {
		if(str->chr[i]!=strw->chr[i]) return 0 ;
		i ++ ;
	}
	return 1 ;
}


unsigned long StrgCharString(Strg str, char * out, unsigned long max) {
	register unsigned long i = 0 ;
	if(str==NULL) {
		errno = EINVAL ;
		DEBUG("StrgShow",1,"NULL Strg") ;
		return -1 ;
	}
	if(out==NULL) {
		return str->len ;
	}
	if(max==0) { out[0] = 0 ; return 0 ; }
	while((i<str->len)&&(i<max)) {
		out[i] = str->chr[i] ;
		i ++ ;
	}
	out[i] = 0 ;
	return i ;
}




//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Lists manipulation
/////


// LIST Object Type
//struct S_Node {
//	struct S_Node * last ;
//	struct S_Node * next ;
//	struct S_Object * obj ;
//} ;
//struct S_List {
//	struct S_Node * frst ;
//	struct S_Node * last ;
//} ;
//typedef struct S_List * List ;
//#define ST_LIST         sizeof(struct S_List)




List ListAlloc() {
	List lst ;
	lst = (List) malloc(sizeof(struct S_List)) ;
	if(lst==NULL) {
		DEBUG("ListAlloc",0,NULL) ;
		return NULL ;
	}
	lst->frst = NULL ;
	lst->last = NULL ;
	return lst ;
}


List ListAllocBlock(unsigned long numb) {
	List lst ;
	struct S_Node * nod ;
	register long i = 1 ;
	lst = (List) malloc(sizeof(struct S_List)) ;
	if(lst==NULL) {
		DEBUG("ListAllocBlock",0,NULL) ;
		return NULL ;
	}
	lst->frst = NULL ;
	lst->last = NULL ;
	if(numb==0) return lst ;
	nod = (struct S_Node *) malloc(sizeof(struct S_Node)) ;
	if(nod==NULL) {
		DEBUG("ListAllocBlock",0,NULL) ;
		return NULL ; 
	}
	nod->last = NULL ;
	nod->obj = NULL ;
	lst->frst = nod ;
	while(i<numb) {
		nod->next = (struct S_Node *) malloc(sizeof(struct S_Node)) ;
		if(nod->next==NULL) {
			DEBUG("ListAllocBlock",0,NULL) ;
			return NULL ;
		}
		nod->next->last = nod ;
		nod = nod->next ;
		nod->obj = NULL ;
		i ++ ;
	}
	nod->next = NULL ;
	lst->last = nod ;
	return lst ;
}


List ListCopy(List lst) {
	List lstl ;
	struct S_Node * nod ;
	struct S_Node * nodl ;
	if(lst==NULL) {
		errno = EINVAL ;
		DEBUG("ListCopy",1,"NULL List") ;
		return NULL ;
	}
	lstl = (List) malloc(sizeof(struct S_List)) ;
	if(lstl==NULL) {
		DEBUG("ListCopy",0,NULL) ;
		return NULL ;
	}
	lstl->frst = NULL ;
	lstl->last = NULL ;
	if( (lst->frst==NULL) || (lst->last==NULL) ) {
		return lstl ;
	}
	nod = lst->frst ;
	nodl = (struct S_Node *) malloc(sizeof(struct S_Node)) ;
	if(nodl==NULL) {
		DEBUG("ListCopy",0,NULL) ;
		return NULL ; 
	}
	nodl->last = NULL ;
	nodl->obj = ObjectCopy(nod->obj) ;
	lstl->frst = nodl ;
	while( (nod->next!=NULL) && (nod!=lst->last) ) {
		nod = nod->next ;
		nodl->next = (struct S_Node *) malloc(sizeof(struct S_Node)) ;
		if(nodl->next==NULL) {
			DEBUG("ListCopy",0,NULL) ;
			return NULL ;
		}
		nodl->next->last = nodl ;
		nodl = nodl->next ;
		nodl->obj = ObjectCopy(nod->obj) ;
	}
	nodl->next = NULL ;
	lstl->last = nodl ;
	return lstl ;
}


List ListClone(List lst) {
	List lstl ;
	struct S_Node * nod ;
	struct S_Node * nodl ;
	if(lst==NULL) {
		errno = EINVAL ;
		DEBUG("ListClone",1,"NULL List") ;
		return NULL ;
	}
	lstl = (List) malloc(sizeof(struct S_List)) ;
	if(lstl==NULL) {
		DEBUG("ListClone",0,NULL) ;
		return NULL ;
	}
	lstl->frst = NULL ;
	lstl->last = NULL ;
	if( (lst->frst==NULL) || (lst->last==NULL) ){
		return lstl ;
	}
	nod = lst->frst ;
	nodl = (struct S_Node *) malloc(sizeof(struct S_Node)) ;
	if(nodl==NULL) {
		DEBUG("ListAllocBlock",0,NULL) ;
		return NULL ; 
	}
	nodl->last = NULL ;
	nodl->obj = nod->obj ;
	lstl->frst = nodl ;
	while( (nod->next!=NULL) && (nod!=lst->last) ) {
		nod = nod->next ;
		nodl->next = (struct S_Node *) malloc(sizeof(struct S_Node)) ;
		if(nodl->next==NULL) {
			DEBUG("ListClone",0,NULL) ;
			return NULL ;
		}
		nodl->next->last = nod ;
		nodl = nodl->next ;
		nodl->obj = nod->obj ;
	}
	nodl->next = NULL ;
	lstl->last = nodl ;
	return lstl ;
}


List ListDup(List lst) {
	struct S_Node * nod ;
	if(lst==NULL) {
		errno = EINVAL ;
		DEBUG("ListDup",1,"NULL List") ;
		return NULL ;
	}
	if( (lst->frst==NULL) || (lst->last==NULL) ) return lst ;
	nod = lst->frst ;
	ObjectDup(nod->obj) ;
	while( (nod->next!=NULL) && (nod!=lst->last) ) {
		nod = nod->next ;
		ObjectDup(nod->obj) ;
	}
	return lst ;
}


List ListDrop(List lst) {
	struct S_Node * nod ;
	if(lst==NULL) {
		errno = EINVAL ;
		DEBUG("ListDrop",1,"NULL List") ;
		return NULL ;
	}
	if( (lst->frst==NULL) || (lst->last==NULL) ) return lst ;
	nod = lst->frst ;
	ObjectDrop(nod->obj) ;
	while( (nod->next!=NULL) && (nod!=lst->last) ) {
		nod = nod->next ;
		ObjectDrop(nod->obj) ;
	}
	return lst ;
}


void ListFree(List lst) {
	struct S_Node * nod ;
	struct S_Node * nodl ;
	if(lst!=NULL) {
		if(lst->frst!=NULL) {
			nod = lst->frst ;
			nod->last = NULL ;
			while( (nod->next!=NULL) && (nod!=lst->last) ) {
				nodl = nod ;
				nod = nod->next ;
				nodl->next = NULL ;
				nodl->last = NULL ;
				ObjectFree(nodl->obj) ;
				nodl->obj = NULL ;
				free(nodl) ;
			}
			if(nod->next!=NULL) nod->next->last = NULL ;
			nod->next = NULL ;
			nod->last = NULL ;
			ObjectFree(nod->obj) ;
			nod->obj = NULL ;
			free(nod) ;
		}
		lst->frst = NULL ;
		lst->last = NULL ;
		free(lst) ;
	}
}


List ListNew() {
	List lst ;
	lst = (List) malloc(sizeof(struct S_List)) ;
	if(lst==NULL) {
		DEBUG("ListNew",0,NULL) ;
		return NULL ;
	}
	lst->frst = NULL ;
	lst->last = NULL ;
	return lst ;
}


Object ListWrap(List lst) {
	Object obj ;
	if(lst==NULL) {
		errno = EINVAL ;
		DEBUG("ListWrap",1,"NULL List") ;
		return NULL ;
	}
	obj = (Object) malloc(sizeof(struct S_Object)) ;
	if(obj==NULL) {
		DEBUG("ListWrap",0,NULL) ;
		return NULL ;
	}
	obj->type = OT_LIST ;
	obj->refc = 0 ;
	obj->flag = 0 ;
	obj->size = ST_LIST ;
	obj->obj = (void *) lst ;
	return obj ;
}


unsigned long ListLen(List lst) {
	struct S_Node * nod ;
	register unsigned long i ;
	if(lst==NULL) {
		errno = EINVAL ;
		DEBUG("ListLen",1,"NULL List") ;
		return -1 ;
	}
	if( (lst->frst==NULL) || (lst->last==NULL) ) return 0 ;
	nod = lst->frst ;
	i = 1 ;
	while( (nod->next!=NULL) && (nod!=lst->last) ) {
		nod = nod->next ;
		i ++ ;
	}
	return i ;
}


Object ListGet(List lst, unsigned long pos) {
	struct S_Node * nod ;
	register unsigned long i = 0 ;
	if(lst==NULL) {
		errno = EINVAL ;
		DEBUG("ListGet",1,"NULL List") ;
		return NULL ;
	}
	if( (lst->frst==NULL) || (lst->last==NULL) ) {
		errno = EINVAL ;
		DEBUG("ListGet",1,"Index out of range") ;
		return NULL ;
	}
	nod = lst->frst ;
	while(i<pos) {
		if( (nod->next==NULL) || (nod==lst->last) ) {
			errno = EINVAL ;
			DEBUG("ListGet",1,"Index out of range") ;
			return NULL ;
		}
		nod = nod->next ;
		i ++ ;
	}
	return nod->obj ;
}


Object ListPut(List lst, unsigned long pos, Object obj) {
	struct S_Node * nod ;
	register unsigned long i = 0 ;
	Object old ;
	if(lst==NULL) {
		errno = EINVAL ;
		DEBUG("ListPut",1,"NULL List") ;
		return NULL ;
	}
	if( (lst->frst==NULL) || (lst->last==NULL) ) {
		errno = EINVAL ;
		DEBUG("ListPut",1,"Index out of range") ;
		return NULL ;
	}
	nod = lst->frst ;
	while(i<pos) {
		if( (nod->next==NULL) || (nod==lst->last) ) {
			errno = EINVAL ;
			DEBUG("ListPut",1,"Index out of range") ;
			return NULL ;
		}
		nod = nod->next ;
		i ++ ;
	}
	old = nod->obj ;
	nod->obj = obj ;
	return old ;
}


List ListDel(List lst, unsigned long beg, unsigned long end) {
	List lstc ;
	struct S_Node * nodb ;
	struct S_Node * nod ;
	register unsigned long i ;
	if(lst==NULL) {
		errno = EINVAL ;
		DEBUG("ListDel",1,"NULL List") ;
		return NULL ;
	}
	lstc = (List) malloc(sizeof(struct S_List)) ;
	if(lstc==NULL) {
		DEBUG("ListDel",0,NULL) ;
		return NULL ;
	}
	lstc->frst = NULL ;
	lstc->last = NULL ;
	if(end<=beg) return lstc ;
	if( (lst->frst==NULL) || (lst->last==NULL) ) {
		free(lstc) ;
		errno = EINVAL ;
		DEBUG("ListDel",1,"Index out of range") ;
		return NULL ;
	}
	if(beg==0) {
		nod = lst->frst ;
		lstc->frst = nod ; // the first deleted one
		i = 1 ;
		while(i<end) {
			if( (nod->next==NULL) || (nod==lst->last) ) {
				free(lstc) ;
				errno = EINVAL ;
				DEBUG("ListDel",1,"Index out of range") ;
				return NULL ;
			}
			nod = nod->next ;
			i ++ ;  
		}
		lstc->last = nod ; // the last deleted one
		if( (nod->next==NULL) || (nod==lst->last) ) {
			lst->frst = NULL ; // resulting list is empty
			lst->last = NULL ;
		}
		else {
			lst->frst = nod->next ; // the first kept one
			lstc->last->next = NULL ; // cutting forward at end of deleted section
			lst->frst->last = NULL ;  // cutting backward at beg of kept section
		}
	}

	else {
		nod = lst->frst ;
		i = 1 ;
		while(i<beg) {
			if( (nod->next==NULL) || (nod==lst->last) ) {
				free(lstc) ;
				errno = EINVAL ;
				DEBUG("ListDel",1,"Index out of range") ;
				return NULL ;
			}
			nod = nod->next ;
			i ++ ;
		}
		nodb = nod ;  // the last one before the deleted section, i = beg >= 1
		if( (nod->next==NULL) || (nod==lst->last) ) {
			errno = EINVAL ;
			DEBUG("ListDel",1,"Index out of range") ;
			return NULL ;
		}
		nod = nod->next ;
		lstc->frst = nod ; // the first deleted one
		i ++ ;
		while(i<end) {
			if( (nod->next==NULL) || (nod==lst->last) ) {
				free(lstc) ;
				errno = EINVAL ;
				DEBUG("ListDel",1,"Index out of range") ;
				return NULL ;
			}
			nod = nod->next ;
			i ++ ;
		}
		lstc->last = nod ; // the last deleted one
		if( (nod->next==NULL) || (nod==lst->last) ) {
			lst->last = nodb ;
			lst->last->next = NULL ; // cutting forward at end of kept section
		}
		else {
			nodb->next = nod->next ; // splicing forward at both cuts
			nodb->next->last = nodb ; // splicing backward
		}
		lstc->frst->last = NULL ; // cutting backward at beg of deleted section
		lstc->last->next = NULL ; // cutting forward at end of deleted section
	}

	return lstc ;
}


List ListIns(List lst, unsigned long pos, List lsi) {
	struct S_Node * nodl ;
	struct S_Node * nod ;
	register unsigned long i = 1 ;
	if( (lst==NULL) || (lsi==NULL) ) {
		errno = EINVAL ;
		DEBUG("ListIns",1,"NULL List") ;
		return NULL ;
	}

	if( (lsi->frst==NULL) || (lsi->last==NULL) ) return lst ;
	nod = lsi->frst ;
	while( (nod->next!=NULL) && (nod!=lsi->last) ) {
		nod = nod->next ;
	} // so nod is the last node of lsi

	if(pos==0) {
		if( (lst->frst==NULL) || (lst->last==NULL) ) {
			nod->next = NULL ;
			lst->last = nod ;
		}
		else {
			nod->next = lst->frst ;
			nod->next->last = nod ;
		}
		lst->frst = lsi->frst ;
		lst->frst->last = NULL ;
	}
	else {
		nodl = lst->frst ;
		while(i<pos) {
			if( (nodl->next==NULL) || (nodl==lst->last) ) {
				errno = EINVAL ;
				DEBUG("ListIns",1,"Index out of range") ;
				return NULL ;
			}
			nodl = nodl->next ;
			i ++ ;
		}
		if( (nodl->next==NULL) || (nodl==lst->last) ) {
			nod->next = NULL ;
			lst->last = nod ;
		}
		else {
			nod->next = nodl->next ;
			nod->next->last = nod ;
		}
		nodl->next = lsi->frst ;
		nodl->next->last = nodl ;
	}
	lsi->frst = NULL ;
	lsi->last = NULL ;
	return lst ;
}


List ListPush(List lst, Object obj) {
	struct S_Node * nod ;
	if(lst==NULL) {
		errno = EINVAL ;
		DEBUG("ListPush",1,"NULL List") ;
		return NULL ;
	}
	if(obj==NULL) {
		errno = EINVAL ;
		DEBUG("ListPush",1,"NULL Object") ;
		return NULL ;
	}
	nod = (struct S_Node *) malloc(sizeof(struct S_Node)) ;
	if(nod==NULL) {
		DEBUG("ListPush",0,NULL) ;
		return NULL ;
	}
	nod->obj = obj ;
	nod->next = NULL ;
	if( (lst->frst==NULL) || (lst->last==NULL) ) {
		lst->frst = nod ;
		nod->last = NULL ;
	}
	else {
		nod->last = lst->last ;
		nod->last->next = nod ;
	}
	lst->last = nod ;
	return lst ;
}


Object ListPop(List lst) {
	struct S_Node * nod ;
	Object obj ;
	if(lst==NULL) {
		errno = EINVAL ;
		DEBUG("ListPop",1,"NULL List") ;
		return NULL ;
	}
	if( (lst->frst==NULL) || (lst->last==NULL) ) {
		errno = EINVAL ;
		DEBUG("ListPop",1,"Empty List") ;
		return NULL ;
	}
	nod = lst->last ;
	obj = nod->obj ;
	lst->last = nod->last ;
	lst->last->next = NULL ;
	free(nod) ;
	return obj ;
}


List ListUnshift(List lst, Object obj) {
	struct S_Node * nod ;
	if(lst==NULL) {
		errno = EINVAL ;
		DEBUG("ListUnshift",1,"NULL List") ;
		return NULL ;
	}
	if(obj==NULL) {
		errno = EINVAL ;
		DEBUG("ListUnshift",1,"NULL Object") ;
		return NULL ;
	}
	nod = (struct S_Node *) malloc(sizeof(struct S_Node)) ;
	if(nod==NULL) {
		DEBUG("ListPush",0,NULL) ;
		return NULL ;
	}
	nod->obj = obj ;
	nod->last = NULL ;
	if( (lst->frst==NULL) || (lst->last==NULL) ) {
		lst->last = nod ;
		nod->next = NULL ;
	}
	else {
		nod->next = lst->frst ;
		nod->next->last = nod ;
	}
	lst->frst = nod ;
	return lst ;
}


Object ListShift(List lst) {
	struct S_Node * nod ;
	Object obj ;
	if(lst==NULL) {
		errno = EINVAL ;
		DEBUG("ListShift",1,"NULL List") ;
		return NULL ;
	}
	if( (lst->frst==NULL) || (lst->last==NULL) ) {
		errno = EINVAL ;
		DEBUG("ListShift",1,"Empty List") ;
		return NULL ;
	}
	nod = lst->frst ;
	obj = nod->obj ;
	lst->frst = nod->next ;
	lst->frst->last = NULL ;
	free(nod) ;
	return obj ;
}


unsigned long ListOut(List lst, char * out, unsigned long max) {
	struct S_Node * nod ;
	register unsigned long i = 0 ;
	if(lst==NULL) {
		errno = EINVAL ;
		DEBUG("ListOut",1,"NULL List") ;
		return -1 ;
	}
	nod = lst->frst ;
	if(out==NULL) {
		i += 3 ;
		if(nod!=NULL) {
			i += ObjectOut(nod->obj,NULL,0) ;
			i ++ ;
			while( (nod->next!=NULL) && (nod!=lst->last) ) {
				nod = nod->next ;
				i += ObjectOut(nod->obj,NULL,0) ;
				i ++ ;
			}
		}
		return i ;
	}
	if(i==max) { out[i] = 0 ; return i ; }
	out[i] = '[' ;
	i ++ ;
	if(i==max) { out[i] = 0 ; return i ; }
	out[i] = ' ' ;
	i ++ ;
	if( (i<max)&&(nod!=NULL) ) {
		i += ObjectOut(nod->obj, out+i, max-i) ;
		if(i==max) { out[i] = 0 ; return i ; }
		out[i] = ' ' ;
		i ++ ;
		while((i<max)&&(nod->next!=NULL)&&(nod!=lst->last)) {
			nod = nod->next ;
			i += ObjectOut(nod->obj, out+i, max-i) ;
			if(i==max) { out[i] = 0 ; return i ; }
			out[i] = ' ' ;
			i ++ ;
		}
	}
	if(i==max) { out[i] = 0 ; return i ; }
	out[i] = ']' ;
	i ++ ;
	out[i] = 0 ;
	return i ;
}


unsigned char ListEqual(List lst, List lstw) {
	struct S_Node * nod ;
	struct S_Node * nodw ;
	char t = 0 ;
	if( (lst==NULL) || (lstw==NULL) ) {
		errno = EINVAL ;
		DEBUG("ListEqual",1,"NULL List") ;
		return 0 ;
	}
	if(lst==lstw) return 1 ;
	if( (lst->frst==NULL) || (lst->last==NULL) ) {
		if( (lstw->frst==NULL) || (lstw->last==NULL) ) {
			return 1 ;
		}
		return 0 ;
	}

	nod = lst->frst ;
	nodw = lstw->frst ;
	if(nod==nodw) return 1 ;
	t = ObjectEqual(nod->obj,nodw->obj) ;
	if(t==-1) return -1 ;
	if(t==0) return 0 ;
	while((nod->next!=NULL)&&(nodw->next!=NULL)&&(nod!=lst->last)&&(nodw!=lstw->last)) {
		nod = nod->next ;
		nodw = nodw->next ;
		if(nod==nodw) return 1 ;
		t = ObjectEqual(nod->obj,nodw->obj) ;
		if(t==-1) return -1 ;
		if(t==0) return 0 ;
	}
	return 0 ;
}




//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Code manipulation
/////



// CODE Object Type
// #define CODE_NAME_LEN 64
// #define CODE_ARGT_LEN  8
// struct S_Code {
//	unsigned char nmcl ;
//	char name[CODE_NAME_LEN+1] ;
//	unsigned char argc ;
//	Type argt[CODE_ARGT_LEN] ;
//	unsigned char retc ;
//	void * func ;
//	struct S_Object * data ;
// } ;
// typedef struct S_Code Code ;
// #define ST_CODE		sizeof(struct S_Code)




Code CodeAlloc() {
	register unsigned long i ;
	Code cod ;
	cod = (Code) malloc(sizeof(struct S_Code)) ;
	if(cod==NULL) {
		DEBUG("CodeAlloc",0,NULL) ;
		return NULL ;
	}
	cod->nmcl = 0 ;
	cod->name[0] = 0 ;
	cod->argc = 0 ;
	cod->retc = 0 ;
	i = 0 ;
	while(i<CODE_ARGT_LEN) {
		cod->argt[i] = 0 ;
		i ++ ;
	}
	cod->func = NULL ;
	cod->data = NULL ;
	return cod ;
}


Code CodeCopy(Code cod) {
	register unsigned long i ;
	Code codl ;
	if(cod==NULL) {
		errno = EINVAL ;
		DEBUG("CodeCopy",1,NULL) ;
		return NULL ;
	}
	codl = (Code) malloc(sizeof(struct S_Code)) ;
	if(codl==NULL) {
		DEBUG("CodeCopy",0,NULL) ;
		return NULL ;
	}
	codl->nmcl = cod->nmcl ;
	i = 0 ;
	while(i<CODE_NAME_LEN) {
		codl->name[i] = cod->name[i] ;
		i ++ ;
	}
	codl->name[i] = 0 ;
	codl->argc = cod->argc ;
	codl->retc = cod->retc ;
	i = 0 ;
	while(i<CODE_ARGT_LEN) {
		codl->argt[i] = cod->argt[i] ;
		i ++ ; 
	}
	codl->func = cod->func ;
	codl->data = ObjectCopy(cod->data) ;
	return codl ;
}


Code CodeClone(Code cod) {
	register unsigned long i ;
	Code codl ;
	if(cod==NULL) {
		errno = EINVAL ;
		DEBUG("CodeClone",1,NULL) ;
		return NULL ;
	}
	codl = (Code) malloc(sizeof(struct S_Code)) ;
	if(codl==NULL) {
		DEBUG("CodeClone",0,NULL) ;
		return NULL ;
	}
	codl->nmcl = cod->nmcl ;
	i = 0 ;
	while(i<CODE_NAME_LEN) {
		codl->name[i] = cod->name[i] ;
		i ++ ;
	}
	codl->name[i] = 0 ;
	codl->argc = cod->argc ;
	codl->retc = cod->retc ;
	i = 0 ;
	while(i<CODE_ARGT_LEN) {
		codl->argt[i] = cod->argt[i] ;
		i ++ ; 
	}
	codl->func = cod->func ;
	codl->data = cod->data ;
	return codl ;
}


Code CodeDup(Code cod) {
	if(cod==NULL) {
		errno = EINVAL ;
		DEBUG("CodeDup",1,"NULL Code") ;
		return NULL ;
	}
	ObjectDup(cod->data) ;
	return cod ;
}


Code CodeDrop(Code cod) {
	if(cod==NULL) {
		errno = EINVAL ;
		DEBUG("CodeDrop",1,"NULL Code") ;
		return NULL ;
	}
	ObjectDrop(cod->data) ;
	return cod ;
}


void CodeFree(Code cod) {
	register unsigned long i ;
	if(cod!=NULL) {
		cod->nmcl = 0 ;
		cod->name[0] = 0 ;
		cod->argc = 0 ;
		cod->retc = 0 ;
		i = 0 ;
		while(i<CODE_ARGT_LEN) {
			cod->argt[i] = 0 ;
			i ++ ;
		}
		cod->func = NULL ;
		ObjectFree(cod->data) ;
		cod->data = NULL ;
		free(cod) ;
	}
}


Code CodeNew() {
	register unsigned long i ;
	Code cod ;
	cod = (Code) malloc(sizeof(struct S_Code)) ;
	if(cod==NULL) {
		DEBUG("CodeNew",0,NULL) ;
		return NULL ;
	}
	cod->nmcl = 0 ;
	cod->name[0] = 0 ;
	cod->argc = 0 ;
	cod->retc = 0 ;
	i = 0 ;
	while(i<CODE_ARGT_LEN) {
		cod->argt[i] = 0 ;
		i ++ ;
	}
	cod->func = NULL ;
	cod->data = NULL ;
	return cod ;
}


Object CodeWrap(Code cod) {
	Object obj ;
	if(cod==NULL) {
		errno = EINVAL ;
		DEBUG("CodeWrap",1,"NULL Code") ;
		return NULL ;
	}
	obj = (Object) malloc(sizeof(struct S_Object)) ;
	if(obj==NULL) {
		DEBUG("CodeWrap",0,NULL) ;
		return NULL ;
	}
	obj->type = OT_CODE ;
	obj->refc = 0 ;
	obj->flag = 0 ;
	obj->size = ST_CODE ;
	obj->obj = (void *) cod ;
	return obj ;
}


unsigned char CodeArgC(Code cod) {
	if(cod==NULL) {
		errno = EINVAL ;
		DEBUG("CodeArgC",1,"NULL Code") ;
		return -1 ;
	}
	return cod->argc ;
}


Type CodeArgT(Code cod, unsigned char argn) {
	if(cod==NULL) {
		errno = EINVAL ;
		DEBUG("CodeArgT",1,"NULL Code") ;
		return -1 ;
	}
	if(argn>=cod->argc) {
		errno = EINVAL ;
		DEBUG("CodeArgT",1,"Invalid index") ;
		return -1 ;
	}
	return cod->argt[argn] ;
}


unsigned char CodeRetC(Code cod) {
	if(cod==NULL) {
		errno = EINVAL ;
		DEBUG("CodeRetC",1,"NULL Code") ;
		return -1 ;
	}
	return cod->retc ;
}


unsigned char CodeSetArgC(Code cod, unsigned char argc) {
	unsigned char oldc ;
	if(cod==NULL) {
		errno = EINVAL ;
		DEBUG("CodeSetArgC",1,"NULL Code") ;
		return -1 ;
	}
	oldc = cod->argc ;
	cod->argc = argc ;
	return oldc ;
}


Type CodeSetArgT(Code cod, unsigned char argn, Type argt) {
	unsigned char oldt ;
	if(cod==NULL) {
		errno = EINVAL ;
		DEBUG("CodeSetArgT",1,"NULL Code") ;
		return -1 ;
	}
	if(argn>=cod->argc) {
		errno = EINVAL ;
		DEBUG("CodeSetArgT",1,"Invalid index") ;
		return -1 ;
	}
	oldt = cod->argt[argn] ;
	cod->argt[argn] = argt ;
	return oldt ;
}


unsigned char CodeSetRetC(Code cod, unsigned char retc) {
	unsigned char oldc ;
	if(cod==NULL) {
		errno = EINVAL ;
		DEBUG("CodeSetRetC",1,"NULL Code") ;
		return -1 ;
	}
	oldc = cod->retc ;
	cod->retc = retc ;
	return oldc ;
}


char * CodeName(Code cod) {
	if(cod==NULL) {
		errno = EINVAL ;
		DEBUG("CodeName",1,"NULL Code") ;
		return NULL ;
	}
	return cod->name ;
}


Strg CodeStrgName(Code cod) {
	Strg str ;
	register unsigned long i = 0 ;
	if(cod==NULL) {
		errno = EINVAL ;
		DEBUG("CodeStrgName",1,"NULL Code") ;
		return NULL ;
	}
	str = (Strg) malloc(sizeof(struct S_Strg)) ;
	if(str==NULL) {
		DEBUG("CodeStrgName",0,NULL) ;
		return NULL ;
	}
	str->chr = (char *) malloc(cod->nmcl) ;
	if(str->chr==NULL) {
		free(str) ;
		DEBUG("CodeStrgName",0,NULL) ;
		return NULL ;
	}
	while(i<cod->nmcl) {
		str->chr[i] = cod->name[i] ;
		i ++ ;
	}
	str->chr[i] = 0 ;
	str->len = cod->nmcl ;
	return str ;
}


void * CodeGetFunc(Code cod) {
	if(cod==NULL) {
		errno = EINVAL ;
		DEBUG("CodeGetFunc",1,"NULL Code") ;
		return NULL ;
	}
	return cod->func ;
}


Object CodeGetData(Code cod) {
	if(cod==NULL) {
		errno = EINVAL ;
		DEBUG("CodeGetData",1,"NULL Code") ;
		return NULL ;
	}
	return cod->data ;
}


void * CodeSetFunc(Code cod, void * func) {
	void * old ;
	if(cod==NULL) {
		errno = EINVAL ;
		DEBUG("CodeSetFunc",1,"NULL Code") ;
		return NULL ;
	}
	old = cod->func ;
	cod->func = func ;
	return old ;
}


Object CodeSetData(Code cod, Object data) {
	Object old ;
	if(cod==NULL) {
		errno = EINVAL ;
		DEBUG("CodeSetData",1,"NULL Code") ;
		return NULL ;
	}
	old = cod->data ;
	cod->data = data ;
	return old ;
}


unsigned long CodeOut(Code cod, char * out, unsigned long max) {
	register unsigned long i = 0 ;
	if(cod==NULL) {
		errno = EINVAL ;
		DEBUG("CodeOut",1,"NULL Code") ;
		return -1 ;
	}
	if(out==NULL) {
		if(cod->data!=NULL) return cod->nmcl+ObjectOut(cod->data,NULL,0)+3 ;
		else return cod->nmcl+2 ;
	}
	if(max==0) { out[0] = 0 ; return 0 ; }
	out[0] = '<' ;
	i = 1 ;
	while((i<=cod->nmcl)&&(i<max)) {
		out[i] = cod->name[i-1] ;
		i ++ ;
	}
	if(i==max) { out[i] = 0 ; return i ; }
	if(cod->data!=NULL) {
		out[i] = ':' ;
		i ++ ;
		if(i==max) { out[i] = 0 ; return i ; }
		i += ObjectOut(cod->data,out+i,max-i) ;
	}
	if(i<max) { out[i] = '>' ; i ++ ; }
	out[i] = 0 ;
	return i ;
}


unsigned char CodeEqual(Code cod, Code codw) {
	register unsigned long i ;
	if((cod==NULL)||(codw==NULL)) {
		errno = EINVAL ;
		DEBUG("CodeEqual",1,"NULL Code") ;
		return -1 ;
	}
	if(cod->nmcl!=codw->nmcl) return  0 ;
	if(cod->argc!=codw->argc) return  0 ;
	if(cod->retc!=codw->retc) return  0 ;
	if(cod->func!=codw->func) return  0 ;
	i = 0 ;
	while(i<CODE_ARGT_LEN) {
		if(cod->argt[i]!=codw->argt[i]) return  0 ;
		i ++ ;
	}
	i = 0 ;
	while(i<cod->nmcl) {
		if(cod->name[i]!=codw->name[i]) return 0 ;
		i ++ ;
	}
	if(!ObjectEqual(cod->data,codw->data)) return 0 ;
	return 1 ;
}






//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Bltn manipulation
/////



// BLTN Object Type
// #define BLTN_NAME_LEN 64
// #define BLTN_ARGT_LEN  8
// struct S_Bltn {
//	char * name ;
//	unsigned char argc ;
//	type argt[BLTN_ARGT_LEN] ;
//	unsigned char retc ;
//	void * func ;
//	char 8 desc ;
// } ;
// typedef struct S_Bltn Bltn ;
// #define ST_BLTN		sizeof(struct S_Bltn)




Bltn BltnAlloc() {
	register unsigned long i ;
	Bltn blt ;
	blt = (Bltn) malloc(sizeof(struct S_Bltn)) ;
	if(blt==NULL) {
		DEBUG("BltnAlloc",0,NULL) ;
		return NULL ;
	}
	blt->name[0] = 0 ;
	blt->argc = 0 ;
	blt->retc = 0 ;
	i = 0 ;
	while(i<BLTN_ARGT_LEN) {
		blt->argt[i] = 0 ;
		i ++ ;
	}
	blt->func = NULL ;
	blt->desc = NULL ;
	return blt ;
}


Bltn BltnCopy(Bltn blt) {
	register unsigned long i ;
	Bltn bltl ;
	if(blt==NULL) {
		errno = EINVAL ;
		DEBUG("BltnCopy",1,NULL) ;
		return NULL ;
	}
	bltl = (Bltn) malloc(sizeof(struct S_Bltn)) ;
	if(bltl==NULL) {
		DEBUG("BltnCopy",0,NULL) ;
		return NULL ;
	}
	i = 0 ;
	while(i<BLTN_NAME_LEN) {
		bltl->name[i] = blt->name[i] ;
		i ++ ;
	}
	bltl->argc = blt->argc ;
	bltl->retc = blt->retc ;
	i = 0 ;
	while(i<BLTN_ARGT_LEN) {
		bltl->argt[i] = blt->argt[i] ;
		i ++ ; 
	}
	bltl->func = blt->func ;
	bltl->desc = blt->desc ;
	return bltl ;
}


Bltn BltnClone(Bltn blt) {
	register unsigned long i ;
	Bltn bltl ;
	if(blt==NULL) {
		errno = EINVAL ;
		DEBUG("BltnClone",1,NULL) ;
		return NULL ;
	}
	bltl = (Bltn) malloc(sizeof(struct S_Bltn)) ;
	if(bltl==NULL) {
		DEBUG("BltnClone",0,NULL) ;
		return NULL ;
	}
	i = 0 ;
	while(i<BLTN_NAME_LEN) {
		bltl->name[i] = blt->name[i] ;
		i ++ ;
	}
	bltl->argc = blt->argc ;
	bltl->retc = blt->retc ;
	i = 0 ;
	while(i<BLTN_ARGT_LEN) {
		bltl->argt[i] = blt->argt[i] ;
		i ++ ; 
	}
	bltl->func = blt->func ;
	bltl->desc = blt->desc ;
	return bltl ;
}


void BltnFree(Bltn blt) {
	if(blt!=NULL) {
		free(blt) ;
	}
}


Bltn BltnNew() {
	register unsigned long i ;
	Bltn blt ;
	blt = (Bltn) malloc(sizeof(struct S_Bltn)) ;
	if(blt==NULL) {
		DEBUG("BltnNew",0,NULL) ;
		return NULL ;
	}
	blt->name[0] = 0 ;
	blt->argc = 0 ;
	blt->retc = 0 ;
	i = 0 ;
	while(i<BLTN_ARGT_LEN) {
		blt->argt[i] = 0 ;
		i ++ ;
	}
	blt->func = NULL ;
	blt->desc = NULL ;
	return blt ;
}


Object BltnWrap(Bltn blt) {
	Object obj ;
	if(blt==NULL) {
		errno = EINVAL ;
		DEBUG("BltnWrap",1,"NULL Bltn") ;
		return NULL ;
	}
	obj = (Object) malloc(sizeof(struct S_Object)) ;
	if(obj==NULL) {
		DEBUG("BltnWrap",0,NULL) ;
		return NULL ;
	}
	obj->type = OT_BLTN ;
	obj->refc = 0 ;
	obj->flag = 0 ;
	obj->size = ST_BLTN ;
	obj->obj = (void *) blt ;
	return obj ;
}


unsigned char BltnArgC(Bltn blt) {
	if(blt==NULL) {
		errno = EINVAL ;
		DEBUG("BltnArgC",1,"NULL Bltn") ;
		return -1 ;
	}
	return blt->argc ;
}


Type BltnArgT(Bltn blt, unsigned char argn) {
	if(blt==NULL) {
		errno = EINVAL ;
		DEBUG("BltnArgT",1,"NULL Bltn") ;
		return -1 ;
	}
	return blt->argt[argn] ;
}


unsigned char BltnRetC(Bltn blt) {
	if(blt==NULL) {
		errno = EINVAL ;
		DEBUG("BltnRetC",1,"NULL Bltn") ;
		return -1 ;
	}
	return blt->retc ;
}


char * BltnName(Bltn blt) {
	if(blt==NULL) {
		errno = EINVAL ;
		DEBUG("BltnName",1,"NULL Bltn") ;
		return NULL ;
	}
	if(blt->name==NULL) {
		return NULL ;
	}
	return blt->name ;
}


Strg BltnStrgName(Bltn blt) {
	register unsigned long i ;
	Strg str ;
	if(blt==NULL) {
		errno = EINVAL ;
		DEBUG("BltnStrgName",1,"NULL Bltn") ;
		return NULL ;
	}
	str = (Strg) malloc(sizeof(struct S_Strg)) ;
	if(str==NULL) {
		DEBUG("BltnStrgName",0,NULL) ;
		return NULL ;
	}
	if(blt->name==NULL) {
		str->len = 0 ;
		str->chr = NULL ;
		return str ;
	}
	i = 0 ;
	while(blt->name[i]!=0) i ++ ;
	str->len = i ;
	str->chr = (char *) malloc(str->len) ;
	if(str->chr==NULL) {
		free(str) ;
		DEBUG("BltnStrgName",0,NULL) ;
		return NULL ;
	}
	while(i<str->len) {
		str->chr[i] = blt->name[i] ;
		i ++ ;
	}
	str->chr[i] = 0 ;
	return str ;
}


void * BltnFunc(Bltn blt) {
	if(blt==NULL) {
		errno = EINVAL ;
		DEBUG("BltnFunc",1,"NULL Bltn") ;
		return NULL ;
	}
	return blt->func ;
}


unsigned long BltnOut(Bltn blt, char * out, unsigned long max) {
	register unsigned long i = 0 ;
	if(blt==NULL) {
		errno = EINVAL ;
		DEBUG("BltnOut",1,"NULL Bltn") ;
		return -1 ;
	}
	if(out==NULL) {
		if(blt->name==NULL) return 2 ;
		while(blt->name[i]!=0) i ++ ;
		return i+2 ;
	}
	if(max==0) { out[0] = 0 ; return 0 ; }
	out[0] = '' ;
	max -- ;
	if(blt->name!=NULL) while(i<max) {
		if(blt->name[i]==0) break ;
		out[i+1] = blt->name[i] ;
		i ++ ;
	}
	if(i<max) { i ++ ; out[i] = '' ; }
	i ++ ;
	out[i] = 0 ;
	return i ;
}


unsigned char BltnEqual(Bltn blt, Bltn bltw) {
	register unsigned long i ;
	if((blt==NULL)||(bltw==NULL)) {
		errno = EINVAL ;
		DEBUG("BltnEqual",1,"NULL Bltn") ;
		return -1 ;
	}
	if(blt->argc!=bltw->argc) return  0 ;
	if(blt->retc!=bltw->retc) return  0 ;
	i = 0 ;
	while(i<BLTN_ARGT_LEN) {
		if(blt->argt[i]!=bltw->argt[i]) return  0 ;
		i ++ ;
	}
	if(blt->func!=bltw->func) return  0 ;
	if(blt->name!=bltw->name) return  0 ;
	return 1 ;
}







