
/*################################################################################################
 *##################                                                            ##################
 *##################                     Parsing module                         ##################
 *##################                                                            ##################
 *##################                     Implementation                         ##################
 *##################                                                            ##################
 *################################################################################################
 */

/*
 * OKit:   Object Tree,  RPL Virtual Machine  and  compiler  shared library
 *
 * Copyright (C) 2000 Yann LANDRIN-SCHWEITZER
 *
 *
 * This library is free software;  you can redistribute it and/or modify it
 * under  the terms of the  GNU Lesser General Public License  as published
 * by the  Free Software Foundation;  either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This  library  is  distributed  in  the hope that it will be useful, but
 * WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty  of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library;  if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * Author:   Yann LANDRIN-SCHWEITZER
 * Contact:  landrin@multimania.com
 * Homepage: http://landrin.multimania.com/
 */



#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>

#include <math.h>

#include <Internal.h>
#include <OKit/ErrLog.h>
#include <OKit/SysTypes.h>
#include <OKit/ObjTypes.h>
#include <OKit/ExecMod.h>

#include <OKit/ParseMod.h>



//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Bltn table query
/////



// Compares "ky1" and "ky2"
// Returns -1, 0, +1 respectively if ky1 is smaller, equal or greater than ky2
signed char _keycmp(char * ky1, char * ky2) {
	register unsigned long i = 0 ;
	while(i<~(unsigned long)0) {
		if(ky1[i]<ky2[i]) return -1 ;
		if(ky1[i]>ky2[i]) return +1 ;
		if(ky1[i]==0) break ;
		i ++ ;
	}
	return 0 ;
}


Bltn * BltnSort(Bltn * table, unsigned long tablesize) {
	Bltn blt ;
	register unsigned long i = 0 ;
	register unsigned long i1 = 0 ;
	register unsigned long i2 = 0 ;
	if(table==NULL) {
		return NULL ;
	}
	if(tablesize==0) {
		return NULL ;
	}
	// heapify the data set
	// we only have to start at the first node with children and
	// work backwards to the top of the heap doing pushdowns
	i = tablesize>>1 ;
	while(i>0) {
		i1 = i ;
		// we will determine the node's max child
		i2 = i1 << 1 ;
		// if this is a leaf node (i.e. it has no children) then we're done
		while(i2<tablesize) {
			// if it has a second child, make i2 the index of the greater kid
			if( ((i2+1)<tablesize) &&
				(_keycmp(table[i2]->name,table[i2+1]->name)==-1) )
				i2 ++ ;
			// now see if the node in question is greater than its max child...
			if(_keycmp(table[i1]->name,table[i2]->name)==-1) {
				// if it's not, swap them and keep going with the push down
				blt = table[i1] ;
				table[i1] = table[i2] ;
				table[i2] = blt ;
				i1 = i2 ;
			}
			else break ;
		}
		i -- ;
	}
	// repeatedly swap top and last then pushdown the promoted last
	i = tablesize-1 ;
	while(i>1) {
		blt = table[1] ;
		table[1] = table[i] ;
		table[i] = blt ;
		i1 = 1 ;
		// we will determine the node's max child
		i2 = i1 << 1 ;
		// if this is a leaf node (i.e. it has no children) then we're done
		while(i2<i) {
			// if it has a second child, make i2 the index of the greater kid
			if( ((i2+1)<i) &&
				(_keycmp(table[i2]->name,table[i2+1]->name)==-1) )
				i2 ++ ;
			// now see if the node in question is greater than its max child...
			if(_keycmp(table[i1]->name,table[i2]->name)==-1) {
				// if it's not, swap them and keep going with the push down
				blt = table[i1] ;
				table[i1] = table[i2] ;
				table[i2] = blt ;
				i1 = i2 ;
			}
			else break ;
		}
		i -- ;
	}
	return table ;
}


Bltn BltnByName(Bltn * table, unsigned long tablesize, char * str) {
	register unsigned long min, max, med ;
	if(table==NULL) {
		return NULL ;
	}
	if(tablesize==0) {
		return NULL ;
	}
	if(str==NULL) {
		return NULL ;
	}
	min = 0 ;
	max = tablesize ;
	switch(_keycmp(str,table[min]->name)) {
		case -1:
			return NULL ;
		case +1:
			break ;
		case 0:
			return table[min] ;
		default:
			return NULL ;
	}
	switch(_keycmp(str,table[max-1]->name)) {
		case -1:
			break ;
		case +1:
			return NULL ;
		case 0:
			return table[max-1] ;
		default:
			return NULL ;
	}
	while(min<max) {
		med = (min+max-1)>>1 ;
		switch(_keycmp(str,table[med]->name)) {
			case -1:
				max = med ;
				break ;
			case +1:
				min = med+1 ;
				break ;
			case 0:
				return table[med] ;
			default:
				return NULL ;
		}
	}
	if(_keycmp(str,table[min]->name)==0)	// min == max here
		return table[min] ;		// ensure we haven't some pathological
						// case such as tablesize == 1
	return NULL ;
}



//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Parsing routines
/////



short isblank(char chr) {
	switch(chr) {
		case 0:    return 1 ;
		case ' ':  return 1 ;
		case '\n': return 1 ;
		case '\t': return 1 ;
		case '\r': return 1 ;
		default:   return 0 ;
	}
	return 0 ;
}

short isdelimiter(char chr) {
	switch(chr) {
		case ',': return 1 ;
		case '': return 1 ;
		case '<': return 1 ;
		case '>': return 1 ;
		case '': return 1 ;
		case '': return 1 ;
		case '[': return 1 ;
		case ']': return 1 ;
		default:   return 0 ;
	}
	return 0 ;
}

short isenddelimiter(char chr) {
	switch(chr) {
		case '>': return 1 ;
		case '': return 1 ;
		case ']': return 1 ;
		default:   return 0 ;
	}
	return 0 ;
}

short isintchar(char chr) {
	switch(chr) {
		case '+': return 1 ;
		case '-': return 1 ;
	}
	if(chr<'0') return 0 ;
	if(chr>'9') return 0 ;
	return 1 ;
}

short isfltchar(char chr) {
	switch(chr) {
		case 'e': return 1 ;
		case 'E': return 1 ;
		case '+': return 1 ;
		case '-': return 1 ;
		case '.': return 1 ;
	}
	if(chr<'0') return 0 ;
	if(chr>'9') return 0 ;
	return 1 ;
}


#define MAX_TOKEN_SIZE (0x1<<20) 

Object ObjectParse(Bltn * table, unsigned long tablesize, char * str, unsigned long * off) {
	register unsigned long i = 0 ;
	unsigned long offv = 0 ;
	char isint ;
	char isflt ;
	char * buffer = NULL ;
	char * tbuffer = NULL ;
	register unsigned long buffsz = 0 ;
	long intval = 0 ;
	long double fltval = 0 ;
	Strg stg ;
	List lst ;
	Bltn blt ;
	Code cod ;
	Object obj ;
	if(str==NULL) {
		errno = EINVAL ;
		DEBUG("ObjectParse",1,"NULL char * pointer") ;
		return NULL ;
	}
	if(off==NULL) off = &offv ;
	while(isblank(str[*off])) {
		if(str[*off]==0) return NULL ;
		(*off) ++ ;
	}
	if(str[*off]=='') {
		(*off) ++ ;
		return VoidWrap(NULL) ;
	}
	if(str[*off]=='<') {
		// A Code Object .. must find it in the table
		(*off) ++ ;
		cod = CodeAlloc() ;
		if(cod==NULL) { return NULL ; }
		i = 0 ;
		while( (i<MAX_TOKEN_SIZE) && (i<CODE_NAME_LEN) ) {
			if(str[*off]==0) {
				CodeFree(cod) ;
				return NULL ;
			}
			if(str[*off]=='>') break ;
			if(str[*off]==':') break ;
			cod->name[i] = str[*off] ;
			(*off) ++ ;
			i ++ ;
		}
		cod->name[i] = 0 ;
		cod->nmcl = i ;
		if(str[*off]==':') { // Looking for Code data
			(*off) ++ ;
			cod->data = ObjectParse(table,tablesize,str,off) ;
			if(str[*off]==0) {
				CodeFree(cod) ;
				return NULL ;
			}
			while(str[*off]!='>') {
				if(!isblank(str[*off])) {
					CodeFree(cod) ;
					return NULL ;
				}
				if(str[*off]==0) {
					CodeFree(cod) ;
					return NULL ;
				}
				(*off) ++ ;
			}
		}
		(*off) ++ ;
		return CodeWrap(cod) ;
	}
	if(str[*off]=='') {
		// A Bltn Object .. must find it in the table
		(*off) ++ ;
		i = 0 ;
		while(i<MAX_TOKEN_SIZE) {
			if(i>=buffsz) {
				tbuffer = realloc(buffer,(buffsz+1)<<1) ;
				if(tbuffer==NULL) {
					free(buffer) ;
					DEBUG("ObjectParse",1,NULL) ;
					return NULL ;
				}
				buffer = tbuffer ;
				buffsz = (buffsz<<1)+1 ;
			}
			if(str[*off]==0) {
				free(buffer) ;
				return NULL ;
			}
			if(str[*off]=='') break ;
			buffer[i] = str[*off] ;
			(*off) ++ ;
			i ++ ;
		}
		(*off) ++ ;
		buffer[i] = 0 ;
		blt = BltnByName(table,tablesize,buffer) ;
		free(buffer) ;
		if(blt==NULL) {
			return NULL ;
		}
		buffer = NULL ;
		return BltnWrap(blt) ;
	}
	if(str[*off]=='[') {
		// A List Object -> recursive parsing
		(*off) ++ ;
		lst = ListAlloc() ;
		if(lst==NULL) {
			return NULL ;
		}
		while(isblank(str[*off])) {
			if(str[*off]==0) return NULL ;
			(*off) ++ ;
		}
		while(str[*off]!=']') {
			obj = ObjectParse(table,tablesize,str,off) ;
			if(obj==NULL) {
				ListFree(lst) ;
				return NULL ;
			}
			if(ListPush(lst,obj)==NULL) {
				ListFree(lst) ;
				return NULL ;
			}
			if(str[*off]==0) {
				ListFree(lst) ;
				return NULL ;
			}
			while(isblank(str[*off])) {
				if(str[*off]==0) return NULL ;
				(*off) ++ ;
			}
		}
		(*off) ++ ;
		return ListWrap(lst) ;
	}
	if(str[*off]=='"') {
		// A Strg Object
		(*off) ++ ;
		i = 0 ;
		while(i<MAX_TOKEN_SIZE) {
			if(i>=buffsz) {
				tbuffer = realloc(buffer,(buffsz<<1)+2) ;
				if(tbuffer==NULL) {
					free(buffer) ;
					DEBUG("ObjectParse",0,NULL) ;
					return NULL ;
				}
				buffer = tbuffer ;
				buffsz = (buffsz<<1)+1 ;
			}
			if(str[*off]==0) {
				free(buffer) ;
				return NULL ;
			}
			if(str[*off]=='"') break ;
			buffer[i] = str[*off] ;
			(*off) ++ ;
			i ++ ;
		}
		(*off) ++ ;
		buffer[i] = 0 ;
		stg = StrgNew(buffer) ;
		free(buffer) ;
		if(stg==NULL) return NULL ;
		obj = StrgWrap(stg) ;
		return obj ;
	}
	// We got a bare word
	isint = 1 ;
	isflt = 1 ;
	i = 0 ;
	while(i<MAX_TOKEN_SIZE) {
		if(i>=buffsz) {
			tbuffer = realloc(buffer,(buffsz<<1)+2) ;
			if(tbuffer==NULL) {
				free(buffer) ;
				DEBUG("ObjectParse",0,NULL) ;
				return NULL ;
			}
			buffer = tbuffer ;
			buffsz = (buffsz<<1)+1 ;
		}
		if(str[*off]==0) break ;
		if(isblank(str[*off])) break ;
		if(isdelimiter(str[*off])) break ;
		if(!isintchar(str[*off])) isint = 0 ;
		if(!isfltchar(str[*off])) isflt = 0 ;
		buffer[i] = str[*off] ;
		(*off) ++ ;
		i ++ ;
	}
	if(i==0) {
		return NULL ;
	}
	buffer[i] = 0 ;
	blt = BltnByName(table,tablesize,buffer) ;
	if(blt!=NULL) {
		free(buffer) ;
		return BltnWrap(blt) ;
	}
	if(isint) {
		Intg in = NULL ;
		switch(buffer[0]) {
			case 0:
				intval = 0 ;
				break ;
			case '-':
				_scaninteger(&intval,buffer+1,i-1,10) ;
				intval = -intval ;
				break ;
			case '+':
				_scaninteger(&intval,buffer+1,i-1,10) ;
				break ;
			default:
				_scaninteger(&intval,buffer,i,10) ;
				break ;
		}
		in = IntgNew(intval) ;
		free(buffer) ;
		if(in==NULL) return NULL ;
		obj = IntgWrap(in) ;
		return obj ;
	}
	if(isflt) {
		Fltp fp = NULL ;
		_scanfloat(&fltval,buffer,i) ;
		fp = FltpNew(fltval) ;
		free(buffer) ;
		if(fp==NULL) return NULL ;
		obj = FltpWrap(fp) ;
		return obj ;
	}
	stg = StrgNew(buffer) ;
	free(buffer) ;
	if(stg==NULL) return NULL ;
	obj = StrgWrap(stg) ;
	return obj ;
}



List ListParse(Bltn * table, unsigned long tablesize, char * str, unsigned long * off) {
	unsigned long offv = 0 ;
	List lst = NULL ;
	Object obj ;
	if(str==NULL) {
		errno = EINVAL ;
		DEBUG("ListParse",1,"NULL char * pointer") ;
		return NULL ;
	}
	if(off==NULL) off = &offv ;
	lst = ListAlloc() ;
	if(lst==NULL) {
		return NULL ;
	}
	while(str[*off]!=0) {
		/*
		if(isenddelimiter(str[*off])) {
			errno = EINVAL ;
			DEBUG("ListParse",1,"Syntax error near \"%10s\" [%u]",str+*off,*off) ;
			ListFree(lst) ;
			return NULL ;
		}
		*/
		obj = ObjectParse(table,tablesize,str,off) ;
		if( (obj==NULL) || (isenddelimiter(str[*off])) ) {
			errno = EINVAL ;
			DEBUG("ListParse",4,"Syntax error near \"%10s\" [%u]",str+*off,*off) ;
			ListFree(lst) ;
			return NULL ;
		}
		if(ListPush(lst,obj)==NULL) {
			ListFree(lst) ;
			return NULL ;
		}
		while(isblank(str[*off])) {
			if(str[*off]==0) return lst ;
			(*off) ++ ;
		}
	}
	return lst ;
}






